""" Module for dealing with VOMS (Virtual Organization Membership Service)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import stat
import tempfile
import shutil

import six
from six.moves import shlex_quote

from DIRAC import S_OK, S_ERROR, gConfig, rootPath, gLogger
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security.ProxyFile import multiProxyArgument, deleteMultiProxy
from DIRAC.Core.Security.BaseSecurity import BaseSecurity
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.Core.Utilities import List, Time, Os


class VOMS(BaseSecurity):
    def __init__(self, timeout=40, *args, **kwargs):
        """Create VOMS class, setting specific timeout for VOMS shell commands."""
        # Per-server timeout for voms-proxy-init, should be at maximum timeout/n
        # where n as the number of voms servers to try.
        self._servTimeout = 12
        super(VOMS, self).__init__(timeout, *args, **kwargs)

    def getVOMSAttributes(self, proxy, switch="all"):
        """
        Return VOMS proxy attributes as list elements if switch="all" (default) OR
        return the string prepared to be stored in DB if switch="db" OR
        return the string of elements to be used as the option string in voms-proxy-init
        if switch="option".
        If a given proxy is a grid proxy, then function will return an empty list.
        """

        # Get all possible info from voms proxy
        result = self.getVOMSProxyInfo(proxy, "all")
        if not result["OK"]:
            return S_ERROR(DErrno.EVOMS, "Failed to extract info from proxy: %s" % result["Message"])

        vomsInfoOutput = List.fromChar(result["Value"], "\n")

        # Get a list of known VOMS attributes
        validVOMSAttrs = []
        result = gConfig.getSections("/Registry/Groups")
        if result["OK"]:
            for group in result["Value"]:
                vA = gConfig.getValue("/Registry/Groups/%s/VOMSRole" % group, "")
                if vA and vA not in validVOMSAttrs:
                    validVOMSAttrs.append(vA)

        # Parse output of voms-proxy-info command
        attributes = []
        voName = ""
        nickName = ""
        for line in vomsInfoOutput:
            fields = List.fromChar(line, ":")
            key = fields[0].strip()
            value = " ".join(fields[1:])
            if key == "VO":
                voName = value
            elif key == "attribute":
                # Cut off unsupported Capability selection part
                if value.find("nickname") == 0:
                    nickName = "=".join(List.fromChar(value, "=")[1:])
                else:
                    value = value.replace("/Capability=NULL", "")
                    value = value.replace("/Role=NULL", "")
                    if value and value not in attributes and value in validVOMSAttrs:
                        attributes.append(value)

        # Sorting and joining attributes
        if switch == "db":
            returnValue = ":".join(attributes)
        elif switch == "option":
            if len(attributes) > 1:
                returnValue = voName + " -order " + " -order ".join(attributes)
            elif attributes:
                returnValue = voName + ":" + attributes[0]
            else:
                returnValue = voName
        elif switch == "nickname":
            returnValue = nickName
        elif switch == "all":
            returnValue = attributes

        return S_OK(returnValue)

    def getVOMSProxyFQAN(self, proxy):
        """Get the VOMS proxy fqan attributes"""
        return self.getVOMSProxyInfo(proxy, "fqan")

    def getVOMSProxyInfo(self, proxy, option=False):
        """
        Returns information about a proxy certificate (both grid and voms).
            Available information is:

              1. Full (grid)voms-proxy-info output
              2. Proxy Certificate Timeleft in seconds (the output is an int)
              3. DN
              4. voms group (if any)

        :type proxy: str
        :param proxy: the proxy certificate location.
        :type  option: str
        :param option: None is the default value. Other option available are:

           * timeleft
           * actimeleft
           * identity
           * fqan
           * all

        :rtype:   tuple
        :return:  status, output, error, pyerror.
        """
        validOptions = ["actimeleft", "timeleft", "identity", "fqan", "all"]
        if option and option not in validOptions:
            return S_ERROR(DErrno.EVOMS, "invalid option %s" % option)

        retVal = multiProxyArgument(proxy)
        if not retVal["OK"]:
            return retVal
        proxyDict = retVal["Value"]

        try:
            res = proxyDict["chain"].getVOMSData()
            if not res["OK"]:
                return res

            data = res["Value"]

            if option == "actimeleft":
                now = Time.dateTime()
                left = data["notAfter"] - now
                return S_OK("%d\n" % left.total_seconds())
            if option == "timeleft":
                now = Time.dateTime()
                left = proxyDict["chain"].getNotAfterDate()["Value"] - now
                return S_OK("%d\n" % left.total_seconds())
            if option == "identity":
                return S_OK("%s\n" % data["subject"])
            if option == "fqan":
                return S_OK(
                    "\n".join([f.replace("/Role=NULL", "").replace("/Capability=NULL", "") for f in data["fqan"]])
                )
            if option == "all":
                lines = []
                creds = proxyDict["chain"].getCredentials()["Value"]
                lines.append("subject : %s" % creds["subject"])
                lines.append("issuer : %s" % creds["issuer"])
                lines.append("identity : %s" % creds["identity"])
                if proxyDict["chain"].isRFC().get("Value"):
                    lines.append("type : RFC compliant proxy")
                else:
                    lines.append("type : proxy")
                left = creds["secondsLeft"]
                h = int(left / 3600)
                m = int(left / 60) - h * 60
                s = int(left) - m * 60 - h * 3600
                lines.append(
                    "timeleft  : %s:%s:%s\nkey usage : Digital Signature, Key Encipherment, Data Encipherment"
                    % (h, m, s)
                )
                lines.append("== VO %s extension information ==" % data["vo"])
                lines.append("VO: %s" % data["vo"])
                lines.append("subject : %s" % data["subject"])
                lines.append("issuer : %s" % data["issuer"])
                for fqan in data["fqan"]:
                    lines.append("attribute : %s" % fqan)
                if "attribute" in data:
                    lines.append("attribute : %s" % data["attribute"])
                now = Time.dateTime()
                left = (data["notAfter"] - now).total_seconds()
                h = int(left / 3600)
                m = int(left / 60) - h * 60
                s = int(left) - m * 60 - h * 3600
                lines.append("timeleft : %s:%s:%s" % (h, m, s))

                return S_OK("\n".join(lines))
            else:
                return S_ERROR(DErrno.EVOMS, "NOT IMP")

        finally:
            if proxyDict["tempFile"]:
                self._unlinkFiles(proxyDict["file"])

    def getVOMSESLocation(self):
        # 755
        requiredDirPerms = stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
        # 644
        requiredFilePerms = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        # 777
        allPerms = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO
        # Transition code to new behaviour
        if "DIRAC_VOMSES" not in os.environ and "X509_VOMSES" not in os.environ:
            os.environ["X509_VOMSES"] = os.path.join(rootPath, "etc", "grid-security", "vomses")
            gLogger.notice(
                "You did not set X509_VOMSES in your bashrc. DIRAC searches $DIRAC/etc/grid-security/vomses . "
                "Please use X509_VOMSES, this auto discovery will be dropped."
            )
        elif "DIRAC_VOMSES" in os.environ and "X509_VOMSES" in os.environ:
            os.environ["X509_VOMSES"] = "%s:%s" % (os.environ["DIRAC_VOMSES"], os.environ["X509_VOMSES"])
            gLogger.notice(
                "You set both variables DIRAC_VOMSES and X509_VOMSES in your bashrc. "
                "DIRAC_VOMSES will be dropped in a future version, please use only X509_VOMSES"
            )
        elif "DIRAC_VOMSES" in os.environ and "X509_VOMSES" not in os.environ:
            os.environ["X509_VOMSES"] = os.environ["DIRAC_VOMSES"]
            gLogger.notice(
                "You set the variables DIRAC_VOMSES in your bashrc. "
                "DIRAC_VOMSES will be dropped in a future version, please use X509_VOMSES"
            )
        # End of transition code
        if "X509_VOMSES" not in os.environ:
            raise Exception(
                "The env variable X509_VOMSES is not set. "
                "DIRAC does not know where to look for etc/grid-security/vomses. "
                "Please set X509_VOMSES in your bashrc."
            )
        vomsesPaths = os.environ["X509_VOMSES"].split(":")
        for vomsesPath in vomsesPaths:
            if not os.path.exists(vomsesPath):
                continue
            if os.path.isfile(vomsesPath):
                pathMode = os.stat(vomsesPath)[stat.ST_MODE]
                if (pathMode & allPerms) ^ requiredFilePerms == 0:
                    return vomsesPath
                fd, tmpPath = tempfile.mkstemp("vomses")
                os.close(fd)
                shutil.copy(vomsesPath, tmpPath)
                os.chmod(tmpPath, requiredFilePerms)
                os.environ["X509_VOMSES"] = tmpPath
                return tmpPath
            elif os.path.isdir(vomsesPath):
                ok = True
                pathMode = os.stat(vomsesPath)[stat.ST_MODE]
                if (pathMode & allPerms) ^ requiredDirPerms:
                    ok = False
                if ok:
                    for fP in os.listdir(vomsesPath):
                        pathMode = os.stat(os.path.join(vomsesPath, fP))[stat.ST_MODE]
                        if (pathMode & allPerms) ^ requiredFilePerms:
                            ok = False
                            break
                if ok:
                    return vomsesPath
                tmpDir = tempfile.mkdtemp()
                tmpDir = os.path.join(tmpDir, "vomses")
                if six.PY2:
                    shutil.copytree(vomsesPath, tmpDir)
                else:
                    # https://bugs.python.org/issue44205
                    shutil.copytree(  # pylint: disable=unexpected-keyword-arg
                        vomsesPath, tmpDir, copy_function=shutil.copy
                    )
                os.chmod(tmpDir, requiredDirPerms)
                for fP in os.listdir(tmpDir):
                    os.chmod(os.path.join(tmpDir, fP), requiredFilePerms)
                os.environ["X509_VOMSES"] = tmpDir
                return tmpDir

    def setVOMSAttributes(self, proxy, attribute=None, vo=None):
        """Sets voms attributes to a proxy"""
        if not vo:
            return S_ERROR(DErrno.EVOMS, "No vo specified, and can't get default in the configuration")

        retVal = multiProxyArgument(proxy)
        if not retVal["OK"]:
            return retVal
        proxyDict = retVal["Value"]
        chain = proxyDict["chain"]
        proxyLocation = proxyDict["file"]

        secs = chain.getRemainingSecs()["Value"] - 300
        if secs < 0:
            return S_ERROR(DErrno.EVOMS, "Proxy length is less that 300 secs")
        hours = int(secs / 3600)
        mins = int((secs - hours * 3600) / 60)

        retVal = self._generateTemporalFile()
        if not retVal["OK"]:
            deleteMultiProxy(proxyDict)
            return retVal
        newProxyLocation = retVal["Value"]

        # DIRACOS 1 only provides a voms-proxy-init2 binary
        cmd = ["voms-proxy-init2" if six.PY2 else "voms-proxy-init"]
        if chain.isLimitedProxy()["Value"]:
            cmd.append("-limited")
        cmd += ["-cert", proxyLocation]
        cmd += ["-key", proxyLocation]
        cmd += ["-out", newProxyLocation]
        cmd += ["-voms"]
        cmd += ["%s:%s" % (vo, attribute) if attribute and attribute != "NoRole" else vo]
        cmd += ["-valid", "%s:%s" % (hours, mins)]
        tmpDir = False
        vomsesPath = self.getVOMSESLocation()
        if vomsesPath:
            cmd += ["-vomses", vomsesPath]

        if chain.isRFC().get("Value"):
            cmd += ["-r"]
        cmd += ["-timeout", str(self._servTimeout)]

        result = shellCall(self._secCmdTimeout, " ".join(map(shlex_quote, cmd)))
        if tmpDir:
            shutil.rmtree(tmpDir)

        deleteMultiProxy(proxyDict)

        if not result["OK"]:
            self._unlinkFiles(newProxyLocation)
            return S_ERROR(DErrno.EVOMS, "Failed to call voms-proxy-init: %s" % result["Message"])

        status, output, error = result["Value"]

        if status:
            self._unlinkFiles(newProxyLocation)
            return S_ERROR(
                DErrno.EVOMS,
                "Failed to set VOMS attributes. Command: %s; StdOut: %s; StdErr: %s" % (cmd, output, error),
            )

        newChain = X509Chain()
        retVal = newChain.loadProxyFromFile(newProxyLocation)
        self._unlinkFiles(newProxyLocation)
        if not retVal["OK"]:
            return S_ERROR(DErrno.EVOMS, "Can't load new proxy: %s" % retVal["Message"])

        return S_OK(newChain)

    def vomsInfoAvailable(self):
        """
        Is voms info available?
        """

        vpInfoCmd = ""
        for vpInfo in ("voms-proxy-info", "voms-proxy-info2"):
            if Os.which(vpInfo):
                vpInfoCmd = vpInfo

        if not vpInfoCmd:
            return S_ERROR(DErrno.EVOMS, "Missing voms-proxy-info")
        cmd = "%s -h" % vpInfoCmd
        result = shellCall(self._secCmdTimeout, cmd)
        if not result["OK"]:
            return False
        status, _output, _error = result["Value"]
        if status:
            return False
        return True



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
from DIRAC.ConfigurationSystem.private.ConfigurationData import ConfigurationData

gConfigurationData = ConfigurationData()



""" Interacts with sqlite3 db
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import sqlite3
import os
import hashlib
import random
import time

import DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.FrameworkSystem.private.monitoring.Activity import Activity
from DIRAC.Core.Utilities import Time


class MonitoringCatalog(object):
    """
    This class is used to perform all kinds queries to the sqlite3 database.
    """

    def __init__(self, dataPath):
        """
        Initialize monitoring catalog
        """
        self.dbConn = False
        self.dataPath = dataPath
        self.log = gLogger.getSubLogger("ActivityCatalog")
        self.createSchema()

    def __connect(self):
        """
        Connects to database
        """
        if not self.dbConn:
            dbPath = "%s/monitoring.db" % self.dataPath
            self.dbConn = sqlite3.connect(dbPath, timeout=20, isolation_level=None)
            # These two settings dramatically increase the performance
            # at the cost of a small corruption risk in case of OS crash
            # It is acceptable though, given the nature of the data
            # details here https://www.sqlite.org/pragma.html
            c = self.dbConn.cursor()
            c.execute("PRAGMA synchronous = OFF")
            c.execute("PRAGMA journal_mode = TRUNCATE")

    def __dbExecute(self, query, values=False):
        """
        Executes a sql statement.

        :type query: string
        :param query: The query to be executed.
        :type values: bool
        :param values: To execute query with values or not.
        :return: the cursor.
        """
        cursor = self.dbConn.cursor()  # pylint: disable=no-member
        self.log.debug("Executing %s" % query)
        executed = False
        retry = 0
        while not executed and retry < 10:
            retry += 1
            try:
                if values:
                    cursor.execute(query, values)
                else:
                    cursor.execute(query)
                executed = True
            except Exception as e:
                self.log.exception("Exception executing statement", "query: %s, values: %s" % (query, values))
                time.sleep(random.random())
        if not executed:
            self.log.error("Could not execute query, big mess ahead", "query: %s, values: %s" % (query, values))
        return cursor

    def __createTables(self):
        """
        Creates tables if not already created
        """
        self.log.info("Creating tables in db")
        try:
            filePath = "%s/monitoringSchema.sql" % os.path.dirname(__file__)
            fd = open(filePath)
            buff = fd.read()
            fd.close()
        except IOError as e:
            DIRAC.abort(1, "Can't read monitoring schema", filePath)
        while buff.find(";") > -1:
            limit = buff.find(";") + 1
            sqlQuery = buff[:limit].replace("\n", "")
            buff = buff[limit:]
            try:
                self.__dbExecute(sqlQuery)
            except Exception as e:
                DIRAC.abort(1, "Can't create tables", str(e))

    def createSchema(self):
        """
        Creates all the sql schema if it does not exist
        """
        self.__connect()
        try:
            sqlQuery = "SELECT name FROM sqlite_master WHERE type='table';"
            c = self.__dbExecute(sqlQuery)
            tablesList = c.fetchall()
            if len(tablesList) < 2:
                self.__createTables()
        except Exception as e:
            self.log.fatal("Failed to startup db engine", str(e))
            return False
        return True

    def __delete(self, table, dataDict):
        """
        Executes an sql delete.

        :type table: string
        :param table: name of the table.
        :type dataDict: dictionary
        :param dataDict: the data dictionary.
        """
        query = "DELETE FROM %s" % table
        valuesList = []
        keysList = []
        for key in dataDict:
            if isinstance(dataDict[key], list):
                orList = []
                for keyValue in dataDict[key]:
                    valuesList.append(keyValue)
                    orList.append("%s = ?" % key)
                keysList.append("( %s )" % " OR ".join(orList))
            else:
                valuesList.append(dataDict[key])
                keysList.append("%s = ?" % key)
        if keysList:
            query += " WHERE %s" % (" AND ".join(keysList))
        self.__dbExecute("%s;" % query, values=valuesList)

    def __select(self, fields, table, dataDict, extraCond="", queryEnd=""):
        """
        Executes a sql select.

        :type fields: string
        :param fields: The fields required in a string.
        :type table: string
        :param table: name of the table.
        :type dataDict: dictionary
        :param dataDict: the data dictionary.
        :return: a list of values.
        """
        valuesList = []
        keysList = []
        for key in dataDict:
            if isinstance(dataDict[key], list):
                orList = []
                for keyValue in dataDict[key]:
                    valuesList.append(keyValue)
                    orList.append("%s = ?" % key)
                keysList.append("( %s )" % " OR ".join(orList))
            else:
                valuesList.append(dataDict[key])
                keysList.append("%s = ?" % key)
        if isinstance(fields, six.string_types):
            fields = [fields]
        if len(keysList) > 0:
            whereCond = "WHERE %s" % (" AND ".join(keysList))
        else:
            whereCond = ""
        if extraCond:
            if whereCond:
                whereCond += " AND %s" % extraCond
            else:
                whereCond = "WHERE %s" % extraCond
        query = "SELECT %s FROM %s %s %s;" % (",".join(fields), table, whereCond, queryEnd)
        c = self.__dbExecute(query, values=valuesList)
        return c.fetchall()

    def __insert(self, table, specialDict, dataDict):
        """
        Executes an sql insert.

        :type table: string
        :param table: name of the table.
        :type specialDict: dictionary
        :param specialDict: the special dictionary.
        :type dataDict: dictionary
        :param dataDict: the data dictionary.
        :return: the number of rows inserted.
        """
        valuesList = []
        valuePoitersList = []
        namesList = []
        for key in specialDict:
            namesList.append(key)
            valuePoitersList.append(specialDict[key])
        for key in dataDict:
            namesList.append(key)
            valuePoitersList.append("?")
            valuesList.append(dataDict[key])
        query = "INSERT INTO %s (%s) VALUES (%s);" % (table, ", ".join(namesList), ",".join(valuePoitersList))
        c = self.__dbExecute(query, values=valuesList)
        return c.rowcount

    def __update(self, newValues, table, dataDict, extraCond=""):
        """
        Executes a sql update.

        :type table: string
        :param table: name of the table.
        :type newValues: dictionary
        :param newValues: a dictionary with new values.
        :type dataDict: dictionary
        :param dataDict: the data dictionary.
        :return: the number of rows updated.
        """
        valuesList = []
        keysList = []
        updateFields = []
        for key in newValues:
            updateFields.append("%s = ?" % key)
            valuesList.append(newValues[key])
        for key in dataDict:
            if isinstance(dataDict[key], list):
                orList = []
                for keyValue in dataDict[key]:
                    valuesList.append(keyValue)
                    orList.append("%s = ?" % key)
                keysList.append("( %s )" % " OR ".join(orList))
            else:
                valuesList.append(dataDict[key])
                keysList.append("%s = ?" % key)
        if len(keysList) > 0:
            whereCond = "WHERE %s" % (" AND ".join(keysList))
        else:
            whereCond = ""
        if extraCond:
            if whereCond:
                whereCond += " AND %s" % extraCond
            else:
                whereCond = "WHERE %s" % extraCond
        query = "UPDATE %s SET %s %s;" % (table, ",".join(updateFields), whereCond)
        c = self.__dbExecute(query, values=valuesList)
        return c.rowcount

    def registerSource(self, sourceDict):
        """
        Registers an activity source.

        :type sourceDict: dictionary
        :param sourceDict: the source dictionary.
        :return: a list of values.
        """
        retList = self.__select("id", "sources", sourceDict)
        if len(retList) > 0:
            return retList[0][0]
        else:
            self.log.info("Registering source", str(sourceDict))
            if self.__insert("sources", {"id": "NULL"}, sourceDict) == 0:
                return -1
            return self.__select("id", "sources", sourceDict)[0][0]

    def registerActivity(self, sourceId, acName, acDict):
        """
        Register an activity.

        :type sourceId: string
        :param sourceId: The source id.
        :type acName: string
        :param acName: name of the activity.
        :type acDict: dictionary
        :param acDict: The activity dictionary containing information about 'category', 'description', 'bucketLength',
                                                                            'type', 'unit'.
        :return: a list of values.
        """
        m = hashlib.md5()
        acDict["name"] = acName
        acDict["sourceId"] = sourceId
        m.update(str(acDict).encode())
        retList = self.__select("filename", "activities", acDict)
        if len(retList) > 0:
            return retList[0][0]
        else:
            acDict["lastUpdate"] = int(Time.toEpoch() - 86000)
            filePath = m.hexdigest()
            filePath = "%s/%s.rrd" % (filePath[:2], filePath)
            self.log.info("Registering activity", str(acDict))
            # This is basically called by the ServiceInterface inside registerActivities method and then all the activity
            # information is stored in the sqlite3 db using the __insert method.

            if (
                self.__insert(
                    "activities",
                    {
                        "id": "NULL",
                        "filename": "'%s'" % filePath,
                    },
                    acDict,
                )
                == 0
            ):
                return -1
            return self.__select("filename", "activities", acDict)[0][0]

    def getFilename(self, sourceId, acName):
        """
        Gets rrd filename for an activity.

        :type sourceId: string
        :param sourceId: The source id.
        :type acName: string
        :param acName: name of the activity.
        :return: The filename in a string.
        """
        queryDict = {"sourceId": sourceId, "name": acName}
        retList = self.__select("filename", "activities", queryDict)
        if len(retList) == 0:
            return ""
        else:
            return retList[0][0]

    def findActivity(self, sourceId, acName):
        """
        Finds activity.

        :type sourceId: string
        :param sourceId: The source id.
        :type acName: string
        :param acName: name of the activity.
        :return: A list containing all the activity information.
        """
        queryDict = {"sourceId": sourceId, "name": acName}
        retList = self.__select(
            "id, name, category, unit, type, description, filename, bucketLength, lastUpdate", "activities", queryDict
        )
        if len(retList) == 0:
            return False
        else:
            return retList[0]

    def activitiesQuery(self, selDict, sortList, start, limit):
        """
        Gets all the sources and activities details in a joined format.

        :type selDict: dictionary
        :param selDict: The fields inside the select query.
        :type sortList: list
        :param sortList: A list in sorted order of the data.
        :type start: int
        :param start: The point or tuple from where to start.
        :type limit: int
        :param limit: The number of tuples to select from the starting point.
        :return: S_OK with a tuple of the result list and fields list.
        """
        fields = [
            "sources.id",
            "sources.site",
            "sources.componentType",
            "sources.componentLocation",
            "sources.componentName",
            "activities.id",
            "activities.name",
            "activities.category",
            "activities.unit",
            "activities.type",
            "activities.description",
            "activities.bucketLength",
            "activities.filename",
            "activities.lastUpdate",
        ]

        extraSQL = ""
        if sortList:
            for sorting in sortList:
                if sorting[0] not in fields:
                    return S_ERROR("Sorting field %s is invalid" % sorting[0])
            extraSQL = "ORDER BY %s" % ",".join(["%s %s" % sorting for sorting in sortList])
        if limit:
            if start:
                extraSQL += " LIMIT %s OFFSET %s" % (limit, start)
            else:
                extraSQL += " LIMIT %s" % limit
        # This method basically takes in some condition and then based on those performs SQL Join on the
        # sources and activities table of the sqlite3 db and returns the corresponding result.
        retList = self.__select(
            ", ".join(fields), "sources, activities", selDict, "sources.id = activities.sourceId", extraSQL
        )
        return S_OK((retList, fields))

    def setLastUpdate(self, sourceId, acName, lastUpdateTime):
        """
        Updates the lastUpdate timestamp for a particular activity using the source id.

        :type sourceId: string
        :param sourceId: The source id.
        :type acName: string
        :param acName: name of the activity.
        :type lastUpdateTime: string
        :param lastUpdateTime: The last update time in the proper format.
        :return: the number of rows updated.
        """
        queryDict = {"sourceId": sourceId, "name": acName}
        return self.__update({"lastUpdate": lastUpdateTime}, "activities", queryDict)

    def getLastUpdate(self, sourceId, acName):
        """
        Gets the lastUpdate timestamp for a particular activity using the source id.

        :type sourceId: string
        :param sourceId: The source id.
        :type acName: string
        :param acName: name of the activity.
        :return: The last update time in string.
        """
        queryDict = {"sourceId": sourceId, "name": acName}
        retList = self.__update("lastUpdate", "activities", queryDict)
        if len(retList) == 0:
            return False
        else:
            return retList[0]

    def queryField(self, field, definedFields):
        """
        Query the values of a field given a set of defined ones.

        :type field: string
        :param field: The field required in a string.
        :type field: list
        :param definedFields: A set of defined fields.
        :return: A list of values.
        """
        retList = self.__select(field, "sources, activities", definedFields, "sources.id = activities.sourceId")
        return retList

    def getMatchingActivities(self, condDict):
        """
        Gets all activities matching the defined conditions.

        :type condDict: dictionary.
        :param condDict: A dictionary containing the conditions.
        :return: a list of matching activities.
        """
        retList = self.queryField(Activity.dbFields, condDict)
        acList = []
        for acData in retList:
            acList.append(Activity(acData))
        return acList

    def registerView(self, viewName, viewData, varFields):
        """
        Registers a new view.

        :type viewName: string
        :param viewName: Name of the view.
        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :type varFields: list
        :param varFields: A list of variable fields.
        :return: S_OK / S_ERROR with the corresponding error message.
        """
        retList = self.__select("id", "views", {"name": viewName})
        if len(retList) > 0:
            return S_ERROR("Name for view name already exists")
        retList = self.__select("name", "views", {"definition": viewData})
        if len(retList) > 0:
            return S_ERROR("View specification already defined with name '%s'" % retList[0][0])
        self.__insert(
            "views", {"id": "NULL"}, {"name": viewName, "definition": viewData, "variableFields": ", ".join(varFields)}
        )
        return S_OK()

    def getViews(self, onlyStatic):
        """
        Gets views.

        :type onlyStatic: bool
        :param onlyStatic: Whether the views required are static or not.
        :return: A list of values.
        """
        queryCond = {}
        if onlyStatic:
            queryCond["variableFields"] = ""
        return self.__select("id, name, variableFields", "views", queryCond)

    def getViewById(self, viewId):
        """
        Gets a view for a given id.

        :type viewId: string
        :param viewId: The view id.
        :return: A list of values.
        """
        if isinstance(viewId, six.string_types):
            return self.__select("definition, variableFields", "views", {"name": viewId})
        else:
            return self.__select("definition, variableFields", "views", {"id": viewId})

    def deleteView(self, viewId):
        """
        Deletes a view for a given id.

        :type viewId: string
        :param viewId: The view id.
        """
        self.__delete("views", {"id": viewId})

    def getSources(self, dbCond, fields=[]):
        """
        Gets souces for a given db condition.

        :type dbCond: dictionary
        :param dbCond: The required database conditions.
        :type fields: list
        :param fields: A list of required fields.
        :return: The list of results after the query is performed.
        """
        if not fields:
            fields = "id, site, componentType, componentLocation, componentName"
        else:
            fields = ", ".join(fields)
        return self.__select(fields, "sources", dbCond)

    def getActivities(self, dbCond):
        """
        Gets activities given a db condition.

        :type dbCond: dictionary
        :param dbCond: The required database conditions.
        :return: a list of activities.
        """
        return self.__select("id, name, category, unit, type, description, bucketLength", "activities", dbCond)

    def deleteActivity(self, sourceId, activityId):
        """
        Deletes an activity.

        :type sourceId: string
        :param sourceId: The source id.
        :type activityId: string
        :param activityId: The activity id.
        :return: S_OK with rrd filename / S_ERROR with a message.
        """
        acCond = {"sourceId": sourceId, "id": activityId}
        acList = self.__select("filename", "activities", acCond)
        if len(acList) == 0:
            return S_ERROR("Activity does not exist")
        rrdFile = acList[0][0]
        self.__delete("activities", acCond)
        acList = self.__select("id", "activities", {"sourceId": sourceId})
        if len(acList) == 0:
            self.__delete("sources", {"id": sourceId})
        return S_OK(rrdFile)



########################################################################
# File: StorageElementHandler.py
########################################################################
"""
:mod: StorageElementHandler

.. module: StorageElementHandler
  :synopsis: StorageElementHandler is the implementation of a simple StorageElement
  service in the DISET framework

The handler implements also the DISET data transfer calls
toClient(), fromClient(), bulkToClient(), bulkFromClient
which support single file, directory and file list upload and download

The class can be used as the basis for more advanced StorageElement implementations

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN StorageElement
  :end-before: ##END
  :dedent: 2
  :caption: StorageElementHandler options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# imports
import os
import shutil
import stat
import re
import errno
import shlex
import six

# from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import mkDir, convertSizeUnits
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.Utilities.Os import getDirectorySize
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Core.Utilities.Adler import fileAdler

from DIRAC.Resources.Storage.StorageBase import StorageBase


BASE_PATH = "/"  # This is not a constant and it will be set in initializeStorageElementHandler() function
MAX_STORAGE_SIZE = 0  # This is in bytes, but in MB in the CS
USE_TOKENS = False


def getDiskSpace(path, total=False):
    """
    Returns disk usage of the given path, in Bytes.
    If total is set to true, the total disk space will be returned instead.
    """

    try:
        st = os.statvfs(path)

        if total:
            # return total space
            queriedSize = st.f_blocks
        else:
            # return free space
            queriedSize = st.f_bavail

        result = float(queriedSize * st.f_frsize)

    except OSError as e:
        return S_ERROR(errno.EIO, "Error while getting the available disk space: %s" % repr(e))

    return S_OK(round(result, 4))


def getTotalDiskSpace():
    """Returns the total maximum volume of the SE storage in B. The total volume
    can be limited either by the amount of the available disk space or by the
    MAX_STORAGE_SIZE value.

    :return: S_OK/S_ERROR, Value is the max total volume of the SE storage in B
    """

    global BASE_PATH
    global MAX_STORAGE_SIZE

    result = getDiskSpace(BASE_PATH, total=True)
    if not result["OK"]:
        return result
    totalSpace = result["Value"]
    maxTotalSpace = min(totalSpace, MAX_STORAGE_SIZE) if MAX_STORAGE_SIZE else totalSpace
    return S_OK(maxTotalSpace)


def getFreeDiskSpace():
    """Returns the free disk space still available for writing taking into account
    the total available disk space and the MAX_STORAGE_SIZE limitation

    :return: S_OK/S_ERROR, Value is the free space of the SE storage in B
    """

    global MAX_STORAGE_SIZE
    global BASE_PATH

    result = getDiskSpace(BASE_PATH)  # free
    if not result["OK"]:
        return result
    if not MAX_STORAGE_SIZE:
        return result

    totalFreeSpace = result["Value"]
    result = getDiskSpace(BASE_PATH, total=True)  # total
    if not result["OK"]:
        return result
    totalSpace = result["Value"]
    totalOccupiedSpace = totalSpace - totalFreeSpace
    freeSpace = min(totalSpace, MAX_STORAGE_SIZE) - totalOccupiedSpace
    return S_OK(freeSpace if freeSpace > 0 else 0)


def initializeStorageElementHandler(serviceInfo):
    """Initialize Storage Element global settings"""

    global BASE_PATH
    global USE_TOKENS
    global MAX_STORAGE_SIZE

    BASE_PATH = getServiceOption(serviceInfo, "BasePath", "")
    if not BASE_PATH:
        gLogger.error("Failed to get the base path")
        return S_ERROR("Failed to get the base path")
    mkDir(BASE_PATH)

    USE_TOKENS = getServiceOption(serviceInfo, "UseTokens", USE_TOKENS)
    MAX_STORAGE_SIZE = convertSizeUnits(getServiceOption(serviceInfo, "MaxStorageSize", MAX_STORAGE_SIZE), "MB", "B")

    gLogger.info("Starting DIRAC Storage Element")
    gLogger.info("Base Path: %s" % BASE_PATH)
    gLogger.info("Max size: %d Bytes" % MAX_STORAGE_SIZE)
    gLogger.info("Use access control tokens: " + str(USE_TOKENS))
    return S_OK()


class StorageElementHandler(RequestHandler):
    """
    .. class:: StorageElementHandler

    """

    def __confirmToken(self, token, path, mode):
        """Confirm the access rights for the path in a given mode"""
        # Not yet implemented
        return True

    @staticmethod
    def __checkForDiskSpace(size):
        """Check if the StorageElement can accommodate 'size' volume of data

        :param int size: size of a new file to store in the StorageElement
        :return: S_OK/S_ERROR, Value is boolean flag True if enough space is available
        """
        result = getFreeDiskSpace()
        if not result["OK"]:
            return result
        freeSpace = result["Value"]
        return S_OK(freeSpace > size)

    def __resolveFileID(self, fileID):
        """get path to file for a given :fileID:"""

        port = self.getCSOption("Port", "")
        if not port:
            return ""

        if ":%s" % port in fileID:
            loc = fileID.find(":%s" % port)
            if loc >= 0:
                fileID = fileID[loc + len(":%s" % port) :]

        serviceName = self.serviceInfoDict["serviceName"]
        loc = fileID.find(serviceName)
        if loc >= 0:
            fileID = fileID[loc + len(serviceName) :]

        loc = fileID.find("?=")
        if loc >= 0:
            fileID = fileID[loc + 2 :]

        if fileID.find(BASE_PATH) == 0:
            return fileID
        while fileID and fileID[0] == "/":
            fileID = fileID[1:]
        return os.path.join(BASE_PATH, fileID)

    @staticmethod
    def __getFileStat(path):
        """Get the file stat information"""
        resultDict = {}
        try:
            statTuple = os.stat(path)
        except OSError as x:
            if str(x).find("No such file") >= 0:
                resultDict["Exists"] = False
                return S_OK(resultDict)
            return S_ERROR("Failed to get metadata for %s" % path)

        resultDict["Exists"] = True
        mode = statTuple[stat.ST_MODE]
        resultDict["Type"] = "File"
        resultDict["File"] = True
        resultDict["Directory"] = False
        if stat.S_ISDIR(mode):
            resultDict["Type"] = "Directory"
            resultDict["File"] = False
        resultDict["Directory"] = True
        resultDict["Size"] = statTuple[stat.ST_SIZE]
        resultDict["TimeStamps"] = (statTuple[stat.ST_ATIME], statTuple[stat.ST_MTIME], statTuple[stat.ST_CTIME])
        resultDict["Cached"] = 1
        resultDict["Migrated"] = 0
        resultDict["Lost"] = 0
        resultDict["Unavailable"] = 0
        resultDict["Mode"] = stat.S_IMODE(mode)

        if resultDict["File"]:
            cks = fileAdler(path)
            resultDict["Checksum"] = cks

        resultDict = StorageBase._addCommonMetadata(resultDict)

        return S_OK(resultDict)

    types_exists = [six.string_types]

    def export_exists(self, fileID):
        """Check existence of the fileID"""
        if os.path.exists(self.__resolveFileID(fileID)):
            return S_OK(True)
        return S_OK(False)

    types_getMetadata = [six.string_types]

    def export_getMetadata(self, fileID):
        """Get metadata for the file or directory specified by fileID"""
        return self.__getFileStat(self.__resolveFileID(fileID))

    types_getFreeDiskSpace = []

    @staticmethod
    def export_getFreeDiskSpace():
        """Get the free disk space of the storage element

        :return: S_OK/S_ERROR, Value is the free space on the SE storage in B
        """
        return getFreeDiskSpace()

    types_getTotalDiskSpace = []

    @staticmethod
    def export_getTotalDiskSpace():
        """Get the total disk space of the storage element

        :return: S_OK/S_ERROR, Value is the max total volume of the SE storage in B
        """
        return getTotalDiskSpace()

    types_createDirectory = [six.string_types]

    def export_createDirectory(self, dir_path):
        """Creates the directory on the storage"""
        path = self.__resolveFileID(dir_path)
        gLogger.info("StorageElementHandler.createDirectory: Attempting to create %s." % path)
        if os.path.exists(path):
            if os.path.isfile(path):
                errStr = "Supplied path exists and is a file"
                gLogger.error("StorageElementHandler.createDirectory: %s." % errStr, path)
                return S_ERROR(errStr)
            gLogger.info("StorageElementHandler.createDirectory: %s already exists." % path)
            return S_OK()
        # Need to think about permissions.
        try:
            mkDir(path)
            return S_OK()
        except Exception as x:
            errStr = "Exception creating directory."
            gLogger.error("StorageElementHandler.createDirectory: %s" % errStr, repr(x))
            return S_ERROR(errStr)

    types_listDirectory = [six.string_types, six.string_types]

    def export_listDirectory(self, dir_path, mode):
        """Return the dir_path directory listing"""
        is_file = False
        path = self.__resolveFileID(dir_path)
        if not os.path.exists(path):
            return S_ERROR("Directory %s does not exist" % dir_path)
        elif os.path.isfile(path):
            fname = os.path.basename(path)
            is_file = True
        else:
            dirList = os.listdir(path)

        resultDict = {}
        if mode == "l":
            if is_file:
                result = self.__getFileStat(fname)
                if result["OK"]:
                    resultDict[fname] = result["Value"]
                    return S_OK(resultDict)
                return S_ERROR("Failed to get the file stat info")
            else:
                failed_list = []
                one_OK = False
                for fname in dirList:
                    result = self.__getFileStat(path + "/" + fname)
                    if result["OK"]:
                        resultDict[fname] = result["Value"]
                        one_OK = True
                    else:
                        failed_list.append(fname)
                if failed_list:
                    if one_OK:
                        result = S_ERROR("Failed partially to get the file stat info")
                    else:
                        result = S_ERROR("Failed to get the file stat info")
                    result["FailedList"] = failed_list
                    result["Value"] = resultDict
                else:
                    result = S_OK(resultDict)

                return result
        else:
            return S_OK(dirList)

    def transfer_fromClient(self, fileID, token, fileSize, fileHelper):
        """Method to receive file from clients.
        fileID is the local file name in the SE.
        fileSize can be Xbytes or -1 if unknown.
        token is used for access rights confirmation.
        """
        result = self.__checkForDiskSpace(fileSize)
        if not result["OK"]:
            return S_ERROR("Failed to get available free space")
        elif not result["Value"]:
            return S_ERROR("Not enough disk space")
        file_path = self.__resolveFileID(fileID)

        if "NoCheckSum" in token:
            fileHelper.disableCheckSum()

        try:
            mkDir(os.path.dirname(file_path))
            with open(file_path, "wb") as fd:
                return fileHelper.networkToDataSink(fd, maxFileSize=(MAX_STORAGE_SIZE))
        except Exception as error:
            return S_ERROR("Cannot open to write destination file %s: %s" % (file_path, str(error)))

    def transfer_toClient(self, fileID, token, fileHelper):
        """Method to send files to clients.
        fileID is the local file name in the SE.
        token is used for access rights confirmation.
        """
        file_path = self.__resolveFileID(fileID)
        if "NoCheckSum" in token:
            fileHelper.disableCheckSum()
        result = fileHelper.getFileDescriptor(file_path, "r")
        if not result["OK"]:
            result = fileHelper.sendEOF()
            # check if the file does not really exist
            if not os.path.exists(file_path):
                return S_ERROR("File %s does not exist" % os.path.basename(file_path))
            return S_ERROR("Failed to get file descriptor")

        fileDescriptor = result["Value"]
        result = fileHelper.FDToNetwork(fileDescriptor)
        fileHelper.oFile.close()
        if not result["OK"]:
            return S_ERROR("Failed to get file " + fileID)
        return result

    def transfer_bulkFromClient(self, fileID, token, _ignoredSize, fileHelper):
        """Receive files packed into a tar archive by the fileHelper logic.
        token is used for access rights confirmation.
        """
        result = self.__checkForDiskSpace(10)
        if not result["OK"]:
            return S_ERROR("Failed to get available free space")
        elif not result["Value"]:
            return S_ERROR("Not enough disk space")
        dirName = fileID.replace(".bz2", "").replace(".tar", "")
        dir_path = self.__resolveFileID(dirName)
        res = fileHelper.networkToBulk(dir_path)
        if not res["OK"]:
            gLogger.error("Failed to receive network to bulk.", res["Message"])
            return res
        if not os.path.exists(dir_path):
            return S_ERROR("Failed to receive data")
        try:
            os.chmod(dir_path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)
        except Exception as error:
            gLogger.exception("Could not set permissions of destination directory.", dir_path, error)
        return S_OK()

    def transfer_bulkToClient(self, fileId, token, fileHelper):
        """Send directories and files specified in the fileID.
        The fileID string can be a single directory name or a list of
        colon (:) separated file/directory names.
        token is used for access rights confirmation.
        """
        fileInput = self.__resolveFileID(fileId)
        tmpList = fileInput.split(":")
        tmpList = [os.path.join(BASE_PATH, x) for x in tmpList]
        strippedFiles = []
        compress = False
        for fileID in tmpList:
            if re.search(".bz2", fileID):
                fileID = fileID.replace(".bz2", "")
                compress = True
            fileID = fileID.replace(".tar", "")
            strippedFiles.append(self.__resolveFileID(fileID))
        res = fileHelper.bulkToNetwork(strippedFiles, compress=compress)
        if not res["OK"]:
            gLogger.error("Failed to send bulk to network", res["Message"])
        return res

    types_remove = [six.string_types, six.string_types]

    def export_remove(self, fileID, token):
        """Remove fileID from the storage. token is used for access rights confirmation."""
        return self.__removeFile(self.__resolveFileID(fileID), token)

    def __removeFile(self, fileID, token):
        """Remove one file with fileID name from the storage"""
        filename = self.__resolveFileID(fileID)
        if self.__confirmToken(token, fileID, "x"):
            try:
                os.remove(filename)
                return S_OK()
            except OSError as error:
                if str(error).find("No such file") >= 0:
                    # File does not exist anyway
                    return S_OK()
                return S_ERROR("Failed to remove file %s" % fileID)
        else:
            return S_ERROR("File removal %s not authorized" % fileID)

    types_getDirectorySize = [six.string_types]

    def export_getDirectorySize(self, fileID):
        """Get the size occupied by the given directory"""
        dir_path = self.__resolveFileID(fileID)
        if os.path.exists(dir_path):
            try:
                space = self.__getDirectorySize(dir_path)
                return S_OK(space)
            except Exception as error:
                gLogger.exception("Exception while getting size of directory", dir_path, error)
                return S_ERROR("Exception while getting size of directory")
        else:
            return S_ERROR("Directory does not exists")

    types_removeDirectory = [six.string_types, six.string_types]

    def export_removeDirectory(self, fileID, token):
        """Remove the given directory from the storage"""

        dir_path = self.__resolveFileID(fileID)
        if not self.__confirmToken(token, fileID, "x"):
            return S_ERROR("Directory removal %s not authorized" % fileID)
        else:
            if not os.path.exists(dir_path):
                return S_OK()
            else:
                try:
                    shutil.rmtree(dir_path)
                    return S_OK()
                except Exception as error:
                    gLogger.error("Failed to remove directory", dir_path)
                    gLogger.error(str(error))
                    return S_ERROR("Failed to remove directory %s" % dir_path)

    types_removeFileList = [list, six.string_types]

    def export_removeFileList(self, fileList, token):
        """Remove files in the given list"""
        failed_list = []
        partial_success = False
        for fileItem in fileList:
            result = self.__removeFile(fileItem, token)
            if not result["OK"]:
                failed_list.append(fileItem)
            else:
                partial_success = True

        if not failed_list:
            return S_OK()
        else:
            if partial_success:
                result = S_ERROR("Bulk file removal partially failed")
                result["FailedList"] = failed_list
            else:
                result = S_ERROR("Bulk file removal failed")
            return result

    ###################################################################

    types_getAdminInfo = []

    @staticmethod
    def export_getAdminInfo():
        """Send the storage element administration information"""
        storageDict = {}
        storageDict["BasePath"] = BASE_PATH
        storageDict["MaxCapacity"] = MAX_STORAGE_SIZE
        used_space = getDirectorySize(BASE_PATH)
        stats = os.statvfs(BASE_PATH)
        available_space = convertSizeUnits(stats.f_bsize * stats.f_bavail, "B", "MB")
        allowed_space = convertSizeUnits(MAX_STORAGE_SIZE, "B", "MB") - used_space
        actual_space = min(available_space, allowed_space)
        storageDict["AvailableSpace"] = actual_space
        storageDict["UsedSpace"] = used_space
        return S_OK(storageDict)

    @staticmethod
    def __getDirectorySize(path):
        """Get the total size of the given directory in bytes"""
        comm = "du -sb %s" % path
        result = systemCall(10, shlex.split(comm))
        if not result["OK"] or result["Value"][0]:
            return 0
        output = result["Value"][1]
        size = int(output.split()[0])
        return size



""" DIRAC FileCatalog component representing a flat directory tree
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

__RCSID__ = "$Id$"

import six
import os
import stat

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import stringListToString, intListToString
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryTreeBase import DirectoryTreeBase


class DirectoryFlatTree(DirectoryTreeBase):
    def __init__(self, database=None):
        DirectoryTreeBase.__init__(self, database)
        self.treeTable = "DirectoryInfo"

    def getDirectoryCounters(self):
        req = "SELECT COUNT(*) FROM DirectoryInfo"
        res = self.db._query(req)
        if not res["OK"]:
            return res
        return S_OK({"DirectoryInfo": res["Value"][0][0]})

    def _findDirectories(self, paths, metadata=[]):
        """Find file ID if it exists for the given list of LFNs"""
        successful = {}
        failed = {}
        req = "SELECT DirName,DirID"
        if metadata:
            req = "%s,%s" % (req, intListToString(metadata))
        req = "%s FROM DirectoryInfo WHERE DirName IN (%s)" % (req, stringListToString(paths))
        res = self.db._query(req)
        if not res["OK"]:
            return res
        for tup in res["Value"]:
            dirName = tup[0]
            dirID = tup[1]
            metaDict = {"DirID": dirID}
            metaDict.update(dict(zip(metadata, tup[2:])))
            successful[dirName] = metaDict
        for path in paths:
            if path not in successful:
                failed[path] = "No such file or directory"
        return S_OK({"Successful": successful, "Failed": failed})

    def __findDirs(self, paths, metadata=["DirName"]):
        dirs = {}
        req = "SELECT DirID,%s FROM DirectoryInfo WHERE DirName IN (%s)" % (
            intListToString(metadata),
            stringListToString(paths),
        )
        res = self.db._query(req)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_OK(dirs)
        for tup in res["Value"]:
            dirID = tup[0]
            dirs[dirID] = dict(zip(metadata, tup[1:]))
        return S_OK(dirs)

    def getPathPermissions(self, paths, credDict):
        """Get the permissions for the supplied paths"""
        res = self.db.ugManager.getUserAndGroupID(credDict)
        if not res["OK"]:
            return res
        uid, gid = res["Value"]
        res = self._findDirectories(paths, metadata=["Mode", "UID", "GID"])
        if not res["OK"]:
            return res
        successful = {}
        for dirName, dirDict in res["Value"]["Successful"].items():
            mode = dirDict["Mode"]
            p_uid = dirDict["UID"]
            p_gid = dirDict["GID"]
            successful[dirName] = {}
            if p_uid == uid:
                successful[dirName]["Read"] = mode & stat.S_IRUSR
                successful[dirName]["Write"] = mode & stat.S_IWUSR
                successful[dirName]["Execute"] = mode & stat.S_IXUSR
            elif p_gid == gid:
                successful[dirName]["Read"] = mode & stat.S_IRGRP
                successful[dirName]["Write"] = mode & stat.S_IWGRP
                successful[dirName]["Execute"] = mode & stat.S_IXGRP
            else:
                successful[dirName]["Read"] = mode & stat.S_IROTH
                successful[dirName]["Write"] = mode & stat.S_IWOTH
                successful[dirName]["Execute"] = mode & stat.S_IXOTH
        return S_OK({"Successful": successful, "Failed": res["Value"]["Failed"]})

    def findDir(self, path):
        res = self.__findDirs([path])
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_OK(0)
        return S_OK(list(res["Value"])[0])

    def removeDir(self, path):
        """Remove directory"""
        res = self.findDir(path)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_OK()
        dirID = res["Value"]
        req = "DELETE FROM DirectoryInfo WHERE DirID=%d" % dirID
        return self.db._update(req)

    def makeDirectory(self, path, credDict, status=0):
        """Create a new directory.
        The return value is the dictionary containing all the parameters of the newly created directory
        """

        if path[0] != "/":
            return S_ERROR("Not an absolute path")

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if result["Value"]:
            return S_OK(result["Value"])

        result = self.db.ugManager.getUserAndGroupID(credDict)
        if not result["OK"]:
            return result
        uid, gid = result["Value"]

        res = self.getParent(path)
        if not res["OK"]:
            return res
        parentID = res["Value"]

        req = (
            "INSERT INTO DirectoryInfo (Parent,Status,DirName,UID,GID,Mode,CreationDate,ModificationDate)\
    VALUES (%d,%d,'%s',%d,%d,%d,UTC_TIMESTAMP(),UTC_TIMESTAMP());"
            % (parentID, status, path, uid, gid, self.db.umask)
        )
        result = self.db._update(req)
        if not result["OK"]:
            self.removeDir(path)
            return S_ERROR("Failed to create directory %s" % path)
        return S_OK(result["lastRowId"])

    def makeDir(self, path):
        result = self.findDir(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]
        if dirID:
            return S_OK(dirID)
        names = ["DirName"]
        values = [path]
        result = self.db.insertFields("DirectoryInfo", names, values)
        if not result["OK"]:
            return result
        return S_OK(result["lastRowId"])

    def existsDir(self, path):
        """Check the existence of a directory at the specified path"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({"Exists": False})
        return S_OK({"Exists": True, "DirID": result["Value"]})

    def getParent(self, path):
        """Get the parent ID of the given directory"""
        return self.findDir(os.path.dirname(path))

    def getParentID(self, dirID):
        """Get the ID of the parent of a directory specified by ID"""
        if dirID == 0:
            return S_ERROR("Root directory ID given")
        req = "SELECT Parent FROM DirectoryInfo WHERE DirID=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("No parent found")
        return S_OK(result["Value"][0][0])

    def getDirectoryPath(self, dirID):
        """Get directory name by directory ID"""
        req = "SELECT DirName FROM DirectoryInfo WHERE DirID=%d" % int(dirID)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory with id %d not found" % int(dirID))
        return S_OK(result["Value"][0][0])

    def getDirectoryName(self, dirID):
        """Get directory name by directory ID"""
        result = self.getDirectoryPath(dirID)
        if not result["OK"]:
            return result
        return S_OK(os.path.basename(result["Value"]))

    def getPathIDs(self, path):
        """Get IDs of all the directories in the parent hierarchy"""
        elements = path.split("/")
        pelements = []
        dPath = ""
        for el in elements[1:]:
            dPath += "/" + el
            pelements.append(dPath)

        pathString = ["'" + p + "'" for p in pelements]
        req = "SELECT DirID FROM DirectoryInfo WHERE DirName in (%s) ORDER BY DirID" % ",".join(pathString)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory %s not found" % path)
        return S_OK([x[0] for x in result["Value"]])

    def getChildren(self, path):
        """Get child directory IDs for the given directory"""
        if isinstance(path, six.string_types):
            result = self.findDir(path)
            if not result["OK"]:
                return result
            dirID = result["Value"]
        else:
            dirID = path
        req = "SELECT DirID FROM DirectoryInfo WHERE Parent=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])
        return S_OK([x[0] for x in result["Value"]])



#!/bin/env python
"""
Script to call the DataRecoveryAgent functionality by hand.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"


class Params(object):
    """Collection of Parameters set via CLI switches."""

    def __init__(self):
        self.enabled = False
        self.transID = 0

    def setEnabled(self, _):
        self.enabled = True
        return S_OK()

    def setTransID(self, transID):
        self.transID = int(transID)
        return S_OK()

    def registerSwitches(self):
        Script.registerSwitch("T:", "TransID=", "TransID to Check/Fix", self.setTransID)
        Script.registerSwitch("X", "Enabled", "Enable the changes", self.setEnabled)


@Script()
def main():
    PARAMS = Params()
    PARAMS.registerSwitches()
    Script.parseCommandLine(ignoreErrors=False)

    # Create Data Recovery Agent and run over single transformation.
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    from DIRAC.TransformationSystem.Agent.DataRecoveryAgent import DataRecoveryAgent

    DRA = DataRecoveryAgent("Transformation/DataRecoveryAgent", "Transformation/DataRecoveryAgent")
    DRA.jobStatus = ["Done", "Failed"]
    DRA.enabled = PARAMS.enabled
    TRANSFORMATION = TransformationClient().getTransformations(condDict={"TransformationID": PARAMS.transID})
    if not TRANSFORMATION["OK"]:
        gLogger.error("Failed to find transformation: %s" % TRANSFORMATION["Message"])
        exit(1)
    if not TRANSFORMATION["Value"]:
        gLogger.error("Did not find any transformations")
        exit(1)
    TRANS_INFO_DICT = TRANSFORMATION["Value"][0]
    TRANS_INFO_DICT.pop("Body", None)
    gLogger.notice("Found transformation: %s" % TRANS_INFO_DICT)
    DRA.treatTransformation(PARAMS.transID, TRANS_INFO_DICT)
    exit(0)


if __name__ == "__main__":
    main()



""" Utility class for dealing with MyProxy
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import re
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.Core.Utilities import List
from DIRAC.Core.Security.ProxyFile import multiProxyArgument, deleteMultiProxy
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security.BaseSecurity import BaseSecurity


class MyProxy(BaseSecurity):
    def uploadProxy(self, proxy=False, useDNAsUserName=False):
        """
        Upload a proxy to myproxy service.
          proxy param can be:
            : Default -> use current proxy
            : string -> upload file specified as proxy
            : X509Chain -> use chain
        """
        retVal = multiProxyArgument(proxy)
        if not retVal["OK"]:
            return retVal
        proxyDict = retVal["Value"]
        chain = proxyDict["chain"]
        proxyLocation = proxyDict["file"]

        timeLeft = int(chain.getRemainingSecs()["Value"] / 3600)

        cmdArgs = ["-n"]
        cmdArgs.append('-s "%s"' % self._secServer)
        cmdArgs.append('-c "%s"' % (timeLeft - 1))
        cmdArgs.append('-t "%s"' % self._secMaxProxyHours)
        cmdArgs.append('-C "%s"' % proxyLocation)
        cmdArgs.append('-y "%s"' % proxyLocation)
        if useDNAsUserName:
            cmdArgs.append("-d")
        else:
            retVal = self._getUsername(chain)
            if not retVal["OK"]:
                deleteMultiProxy(proxyDict)
                return retVal
            mpUsername = retVal["Value"]
            cmdArgs.append('-l "%s"' % mpUsername)

        mpEnv = self._getExternalCmdEnvironment()
        # Hack to upload properly
        mpEnv["GT_PROXY_MODE"] = "old"

        cmd = "myproxy-init %s" % " ".join(cmdArgs)
        result = shellCall(self._secCmdTimeout, cmd, env=mpEnv)

        deleteMultiProxy(proxyDict)

        if not result["OK"]:
            errMsg = "Call to myproxy-init failed: %s" % retVal["Message"]
            return S_ERROR(errMsg)

        status, _, error = result["Value"]

        # Clean-up files
        if status:
            errMsg = "Call to myproxy-init failed"
            extErrMsg = "Command: %s; StdOut: %s; StdErr: %s" % (cmd, result, error)
            return S_ERROR("%s %s" % (errMsg, extErrMsg))

        return S_OK()

    def getDelegatedProxy(self, proxyChain, lifeTime=604800, useDNAsUserName=False):
        """
        Get delegated proxy from MyProxy server
        return S_OK( X509Chain ) / S_ERROR
        """
        # TODO: Set the proxy coming in proxyString to be the proxy to use

        # Get myproxy username diracgroup:diracuser
        retVal = multiProxyArgument(proxyChain)
        if not retVal["OK"]:
            return retVal
        proxyDict = retVal["Value"]
        chain = proxyDict["chain"]
        proxyLocation = proxyDict["file"]

        retVal = self._generateTemporalFile()
        if not retVal["OK"]:
            deleteMultiProxy(proxyDict)
            return retVal
        newProxyLocation = retVal["Value"]

        # myproxy-get-delegation works only with environment variables
        cmdEnv = self._getExternalCmdEnvironment()
        if self._secRunningFromTrustedHost:
            cmdEnv["X509_USER_CERT"] = self._secCertLoc
            cmdEnv["X509_USER_KEY"] = self._secKeyLoc
            if "X509_USER_PROXY" in cmdEnv:
                del cmdEnv["X509_USER_PROXY"]
        else:
            cmdEnv["X509_USER_PROXY"] = proxyLocation

        cmdArgs = []
        cmdArgs.append("-s '%s'" % self._secServer)
        cmdArgs.append("-t '%s'" % (int(lifeTime / 3600)))
        cmdArgs.append("-a '%s'" % proxyLocation)
        cmdArgs.append("-o '%s'" % newProxyLocation)
        if useDNAsUserName:
            cmdArgs.append("-d")
        else:
            retVal = self._getUsername(chain)
            if not retVal["OK"]:
                deleteMultiProxy(proxyDict)
                return retVal
            mpUsername = retVal["Value"]
            cmdArgs.append('-l "%s"' % mpUsername)

        cmd = "myproxy-logon %s" % " ".join(cmdArgs)
        gLogger.verbose("myproxy-logon command:\n%s" % cmd)

        result = shellCall(self._secCmdTimeout, cmd, env=cmdEnv)

        deleteMultiProxy(proxyDict)

        if not result["OK"]:
            errMsg = "Call to myproxy-logon failed: %s" % result["Message"]
            deleteMultiProxy(proxyDict)
            return S_ERROR(errMsg)

        status, _, error = result["Value"]

        # Clean-up files
        if status:
            errMsg = "Call to myproxy-logon failed"
            extErrMsg = "Command: %s; StdOut: %s; StdErr: %s" % (cmd, result, error)
            deleteMultiProxy(proxyDict)
            return S_ERROR("%s %s" % (errMsg, extErrMsg))

        chain = X509Chain()
        retVal = chain.loadProxyFromFile(newProxyLocation)
        if not retVal["OK"]:
            deleteMultiProxy(proxyDict)
            return S_ERROR("myproxy-logon failed when reading delegated file: %s" % retVal["Message"])

        deleteMultiProxy(proxyDict)
        return S_OK(chain)

    def getInfo(self, proxyChain, useDNAsUserName=False):
        """
        Get info from myproxy server

        :return: S_OK( { 'username' : myproxyusername,
                       'owner' : owner DN,
                       'timeLeft' : secs left } ) / S_ERROR
        """
        # TODO: Set the proxy coming in proxyString to be the proxy to use

        # Get myproxy username diracgroup:diracuser
        retVal = multiProxyArgument(proxyChain)
        if not retVal["OK"]:
            return retVal
        proxyDict = retVal["Value"]
        chain = proxyDict["chain"]
        proxyLocation = proxyDict["file"]

        # myproxy-get-delegation works only with environment variables
        cmdEnv = self._getExternalCmdEnvironment()
        if self._secRunningFromTrustedHost:
            cmdEnv["X509_USER_CERT"] = self._secCertLoc
            cmdEnv["X509_USER_KEY"] = self._secKeyLoc
            if "X509_USER_PROXY" in cmdEnv:
                del cmdEnv["X509_USER_PROXY"]
        else:
            cmdEnv["X509_USER_PROXY"] = proxyLocation

        cmdArgs = []
        cmdArgs.append("-s '%s'" % self._secServer)
        if useDNAsUserName:
            cmdArgs.append("-d")
        else:
            retVal = self._getUsername(chain)
            if not retVal["OK"]:
                deleteMultiProxy(proxyDict)
                return retVal
            mpUsername = retVal["Value"]
            cmdArgs.append('-l "%s"' % mpUsername)

        cmd = "myproxy-info %s" % " ".join(cmdArgs)
        gLogger.verbose("myproxy-info command:\n%s" % cmd)

        result = shellCall(self._secCmdTimeout, cmd, env=cmdEnv)

        deleteMultiProxy(proxyDict)

        if not result["OK"]:
            errMsg = "Call to myproxy-info failed: %s" % result["Message"]
            deleteMultiProxy(proxyDict)
            return S_ERROR(errMsg)

        status, output, error = result["Value"]

        # Clean-up files
        if status:
            errMsg = "Call to myproxy-info failed"
            extErrMsg = "Command: %s; StdOut: %s; StdErr: %s" % (cmd, result, error)
            return S_ERROR("%s %s" % (errMsg, extErrMsg))

        infoDict = {}
        usernameRE = re.compile(r"username\s*:\s*(\S*)")
        ownerRE = re.compile(r"owner\s*:\s*(\S*)")
        timeLeftRE = re.compile(r"timeleft\s*:\s*(\S*)")
        for line in List.fromChar(output, "\n"):
            match = usernameRE.search(line)
            if match:
                infoDict["username"] = match.group(1)
            match = ownerRE.search(line)
            if match:
                infoDict["owner"] = match.group(1)
            match = timeLeftRE.search(line)
            if match:
                try:
                    fields = List.fromChar(match.group(1), ":")
                    fields.reverse()
                    secsLeft = 0
                    for iP in range(len(fields)):
                        if iP == 0:
                            secsLeft += int(fields[iP])
                        elif iP == 1:
                            secsLeft += int(fields[iP]) * 60
                        elif iP == 2:
                            secsLeft += int(fields[iP]) * 3600
                    infoDict["timeLeft"] = secsLeft
                except Exception as x:
                    print(x)

        return S_OK(infoDict)



""" Collection of utilities function
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib
import os
import time

from DIRAC.Core.Workflow.Module import ModuleDefinition
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.Core.Workflow.Step import StepDefinition

#############################################################################


def getStepDefinition(stepName, modulesNameList=None, importLine="", parametersList=None):
    """Given a name, a list of modules name, and a list of parameters, returns a step definition.
    Remember that Step definition = Parameters + Module Instances
    """
    if modulesNameList is None:
        modulesNameList = []

    if parametersList is None:
        parametersList = []

    stepDef = StepDefinition(stepName)

    for moduleName in modulesNameList:
        module = None
        if importLine:
            try:
                module = importlib.import_module(importLine + "." + moduleName)
            except ImportError:
                pass
        # In case the importLine is not set, this is looking for a DIRAC extension, if any
        if module is None:
            module = ObjectLoader().loadModule("Workflow.Modules." + moduleName)["Value"]

        # create the module definition
        moduleDef = ModuleDefinition(moduleName)
        moduleDef.setDescription(module.__doc__)
        moduleDef.setBody("\nfrom %s import %s\n" % (module.__name__, moduleName))

        # add the module to the step, and instance it
        stepDef.addModule(moduleDef)
        stepDef.createModuleInstance(module_type=moduleName, name=moduleName)

    # add parameters to the module definition
    for pName, pType, pValue, pDesc in parametersList:
        p = Parameter(pName, pValue, pType, "", "", True, False, pDesc)
        stepDef.addParameter(Parameter(parameter=p))

    return stepDef


#############################################################################


def addStepToWorkflow(workflow, stepDefinition, name):
    """Add a stepDefinition to a workflow, instantiating it, and giving it a name"""

    workflow.addStep(stepDefinition)
    return workflow.createStepInstance(stepDefinition.getType(), name)


#############################################################################


def getStepCPUTimes(step_commons):
    """CPU times of a step"""
    exectime = 0
    if "StartTime" in step_commons:
        exectime = time.time() - step_commons["StartTime"]

    cputime = 0
    if "StartStats" in step_commons:
        # os.times() returns a 5-tuple (utime, stime, cutime, cstime, elapsed_time)
        cputimeNow = sum(os.times()[:4])
        cputimeBefore = sum(step_commons["StartStats"][:4])
        cputime = cputimeNow - cputimeBefore

    return exectime, cputime


#############################################################################



""" RPCClient object is used to create RPC connection to services
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


from DIRAC.Core.DISET.private.InnerRPCClient import InnerRPCClient


class _MagicMethod(object):
    """This object allows to bundle together a function calling
    an RPC and the remote function name.
    When this object is called (__call__), the call is performed.
    """

    def __init__(self, doRPCFunc, remoteFuncName):
        """Constructor

        :param doRPCFunc: the function actually performing the RPC call
        :param remoteFuncName: name of the remote function
        """
        self.__doRPCFunc = doRPCFunc
        self.__remoteFuncName = remoteFuncName

    def __getattr__(self, remoteFuncName):
        """I really do not understand when this would be called.
        I can only imagine it being called by dir, or things like that.
        In any case, it recursively return a MagicMethod object
        where the new remote function name is the old one to which
        we append the new called attribute.
        """
        return _MagicMethod(self.__doRPCFunc, "%s.%s" % (self.__remoteFuncName, remoteFuncName))

    def __call__(self, *args, **kwargs):
        """Triggers the call.
        it uses the RPC calling function given by RPCClient,
        and gives as argument the remote function name and whatever
        arguments given.
        """
        return self.__doRPCFunc(self.__remoteFuncName, args, **kwargs)

    def __str__(self):
        return "<RPCClient method %s>" % self.__remoteFuncName


class RPCClient(object):
    """This class contains the mechanism to convert normal calls to RPC calls.

    When instanciated, it creates a :class:`~DIRAC.Core.DISET.private.InnerRPCClient.InnerRPCClient`
    as an attribute. Any attribute which is accessed is then either redirected to InnerRPCClient if it has it,
    or creates a MagicMethod object otherwise. If the attribute is a function, MagicMethod will
    trigger the RPC call, using the InnerRPCClient.

    The typical workflow looks like this::

      rpc = RPCClient('DataManagement/FileCatalog')

      # Here, func is the ping function, which we call remotely.
      # We go through RPCClient.__getattr__ which returns us a MagicMethod object
      func = rpc.ping

      # Here we call the method __call__ of the MagicMethod
      func()

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor
        The arguments are just passed on to InnerRPCClient.
        In practice:

          * args: has to be the service name or URL
          * kwargs: all the arguments InnerRPCClient and BaseClient accept as configuration
        """
        self.__innerRPCClient = InnerRPCClient(*args, **kwargs)

    def __doRPC(self, sFunctionName, args, **kwargs):
        """
        Execute the RPC action. This is given as an attribute
        to MagicMethod

        :param sFunctionName: name of the remote function
        :param args: arguments to pass to the function
        """
        return self.__innerRPCClient.executeRPC(sFunctionName, args, **kwargs)

    def __getattr__(self, attrName):
        """Function for emulating the existence of functions.

          In literature this is usually called a "stub function".
        If the attribute exists in InnerRPCClient, return it,
        otherwise we create a _MagicMethod instance

        """
        if attrName in dir(self.__innerRPCClient):
            return getattr(self.__innerRPCClient, attrName)
        return _MagicMethod(self.__doRPC, attrName)


def executeRPCStub(rpcStub):
    """
    Playback a stub
    """
    # Generate a RPCClient with the same parameters
    rpcClient = RPCClient(rpcStub[0][0], **rpcStub[0][1])
    # Get a functor to execute the RPC call
    rpcFunc = getattr(rpcClient, rpcStub[1])
    # Reproduce the call
    return rpcFunc(*rpcStub[2])



""" ElementInspectorAgent

  This agent inspect Resources (or maybe Nodes), and evaluates policies that apply.


The following options can be set for the ElementInspectorAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN ElementInspectorAgent
  :end-before: ##END
  :dedent: 2
  :caption: ElementInspectorAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import concurrent.futures

from DIRAC import S_ERROR, S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ResourceStatusSystem.PolicySystem.PEP import PEP

AGENT_NAME = "ResourceStatus/ElementInspectorAgent"


class ElementInspectorAgent(AgentModule):
    """ElementInspectorAgent

    The ElementInspector agent is a generic agent used to check the elements
    of type "Resource" -- which includes ComputingElement, StorageElement, and other types

    This Agent takes care of the Elements. In order to do so, it gathers
    the eligible ones and then evaluates their statuses with the PEP.
    """

    # Inspection freqs, defaults, the lower, the higher priority to be checked.
    # Error state usually means there is a glitch somewhere, so it has the highest
    # priority.
    __checkingFreqs = {"Active": 20, "Degraded": 20, "Probing": 20, "Banned": 15, "Unknown": 10, "Error": 5}

    def __init__(self, *args, **kwargs):
        """c'tor"""

        AgentModule.__init__(self, *args, **kwargs)

        # ElementType, to be defined among Resource or Node
        self.elementType = "Resource"
        self.rsClient = None
        self.clients = {}

    def initialize(self):
        """Standard initialize."""

        self.elementType = self.am_getOption("elementType", self.elementType)

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceStatusClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceStatusClient class: %s" % res["Message"])
            return res
        rsClass = res["Value"]

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceManagementClient class: %s" % res["Message"])
            return res
        rmClass = res["Value"]

        self.rsClient = rsClass()
        self.clients["ResourceStatusClient"] = rsClass()
        self.clients["ResourceManagementClient"] = rmClass()

        if not self.elementType:
            return S_ERROR("Missing elementType")

        maxNumberOfThreads = self.am_getOption("maxNumberOfThreads", 15)
        self.log.info("Multithreaded with %d threads" % maxNumberOfThreads)
        self.threadPoolExecutor = concurrent.futures.ThreadPoolExecutor(max_workers=maxNumberOfThreads)

        return S_OK()

    def execute(self):
        """
        This is the main method of the agent.
        It gets the elements from the Database which are eligible to be re-checked.

        Gets all the rows in the <self.elementType>Status table, and then
        discards entries with TokenOwner != rs_svc. On top of that, there are check
        frequencies that are applied: depending on the current status of the element,
        they will be checked more or less often.
        """

        # We get all the elements, then we filter.
        res = self.rsClient.selectStatusElement(self.elementType, "Status")
        if not res["OK"]:
            return res

        utcnow = datetime.datetime.utcnow().replace(microsecond=0)
        future_to_element = {}

        # filter elements by Type
        for element in res["Value"]:

            # Maybe an overkill, but this way I have NEVER again to worry about order
            # of elements returned by mySQL on tuples
            elemDict = dict(zip(res["Columns"], element))

            # This if-clause skips all the elements that should not be checked yet
            timeToNextCheck = self.__checkingFreqs[elemDict["Status"]]
            if utcnow <= elemDict["LastCheckTime"] + datetime.timedelta(minutes=timeToNextCheck):
                continue

            # We skip the elements with token different than "rs_svc"
            if elemDict["TokenOwner"] != "rs_svc":
                self.log.verbose(
                    "Skipping %s ( %s ) with token %s"
                    % (elemDict["Name"], elemDict["StatusType"], elemDict["TokenOwner"])
                )
                continue

            # if we are here, we process the current element
            self.log.verbose(
                '%s # "%s" # "%s" # %s # %s'
                % (
                    elemDict["Name"],
                    elemDict["ElementType"],
                    elemDict["StatusType"],
                    elemDict["Status"],
                    elemDict["LastCheckTime"],
                )
            )
            lowerElementDict = {"element": self.elementType}
            for key, value in elemDict.items():
                if len(key) >= 2:  # VO !
                    lowerElementDict[key[0].lower() + key[1:]] = value
            # We process lowerElementDict
            future = self.threadPoolExecutor.submit(self._execute, lowerElementDict)
            future_to_element[future] = elemDict["Name"]

        for future in concurrent.futures.as_completed(future_to_element):
            transID = future_to_element[future]
            try:
                future.result()
            except Exception as exc:
                self.log.exception("%s generated an exception: %s" % (transID, exc))
            else:
                self.log.info("Processed", transID)

        return S_OK()

    def _execute(self, element):
        """
        Evaluates the policies for an element and enforces the necessary actions.
        """

        pep = PEP(clients=self.clients)

        self.log.verbose(
            "%s ( VO=%s / status=%s / statusType=%s ) being processed"
            % (element["name"], element["vO"], element["status"], element["statusType"])
        )

        try:
            res = pep.enforce(element)
        except Exception:
            self.log.exception("Exception during enforcement")
            res = S_ERROR("Exception during enforcement")
        if not res["OK"]:
            self.log.error("Failed policy enforcement", res["Message"])
            return res

        resEnforce = res["Value"]

        oldStatus = resEnforce["decisionParams"]["status"]
        statusType = resEnforce["decisionParams"]["statusType"]
        newStatus = resEnforce["policyCombinedResult"]["Status"]
        reason = resEnforce["policyCombinedResult"]["Reason"]

        if oldStatus != newStatus:
            self.log.info(
                "%s (%s) is now %s ( %s ), before %s" % (element["name"], statusType, newStatus, reason, oldStatus)
            )

    def finalize(self):
        """graceful finalization"""

        self.log.info("Wait for threads to get empty before terminating the agent")
        self.threadPoolExecutor.shutdown()
        self.log.info("Threads are empty, terminating the agent...")
        return S_OK()



""" Hosts BaseTransport class, which is a base for PlainTransport and SSLTransport

BaseTransport is used by the client and the service, it describes an opened connection.
Here a diagram of basic Client/Service exchange

Client -> ServiceReactor : Connect

Client<-->Service        : Handshake (only in SSLTransport)

Client -> Service        : Propose action

Client <- Service        : S_OK

Client -> RequestHandler : Arguments

Client <- RequestHandler : Response

Client <- Service        : Close
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import six
from six import BytesIO
from hashlib import md5

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from DIRAC.Core.Utilities.ReturnValues import S_ERROR, S_OK
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities import MixedEncode


class BaseTransport(object):
    """Invokes MixedEncode for marshaling/unmarshaling of data calls in transit"""

    bAllowReuseAddress = True

    # This option corresponds to the `backlog` param of the `listen` syscall.
    # you may want to read the man page before tuning it...
    iListenQueueSize = 128
    iReadTimeout = 600
    keepAliveMagic = b"dka"

    def __init__(self, stServerAddress, bServerMode=False, **kwargs):
        self.bServerMode = bServerMode
        self.extraArgsDict = kwargs
        self.byteStream = b""
        self.packetSize = 1048576  # 1MiB
        self.stServerAddress = stServerAddress
        self.peerCredentials = {}
        self.remoteAddress = False
        self.appData = ""
        self.startedKeepAlives = set()
        self.keepAliveId = md5((str(stServerAddress) + str(bServerMode)).encode()).hexdigest()
        self.receivedMessages = []
        self.sentKeepAlives = 0
        self.waitingForKeepAlivePong = False
        self.__keepAliveLapse = 0
        self.oSocket = None
        if "keepAliveLapse" in kwargs:
            try:
                self.__keepAliveLapse = max(150, int(kwargs["keepAliveLapse"]))
            except Exception:
                pass
        self.iListenQueueSize = max(self.iListenQueueSize, int(kwargs.get("SocketBacklog", 0)))
        self.__lastActionTimestamp = time.time()
        self.__lastServerRenewTimestamp = self.__lastActionTimestamp

    def __updateLastActionTimestamp(self):
        self.__lastActionTimestamp = time.time()

    def getLastActionTimestamp(self):
        return self.__lastActionTimestamp

    def getKeepAliveLapse(self):
        return self.__keepAliveLapse

    def handshake(self):
        """This method is overwritten by SSLTransport if we use a secured transport."""
        return S_OK()

    def close(self):
        self.oSocket.close()

    def setAppData(self, appData):
        self.appData = appData

    def getAppData(self):
        return self.appData

    def renewServerContext(self):
        self.__lastServerRenewTimestamp = time.time()
        return S_OK()

    def latestServerRenewTime(self):
        return self.__lastServerRenewTimestamp

    def getConnectingCredentials(self):
        """

        :return: dictionnary with credentials

          Return empty dictionnary for plainTransport.

          In SSLTransport it contains (after the handshake):

           - 'DN' : All identity name, e.g. ```/C=ch/O=DIRAC/OU=DIRAC CI/CN=ciuser```
           - 'CN' : Only the user name e.g. ciuser
           - 'x509Chain' : List of all certificates in the chain
           - 'isProxy' : True if the client use proxy certificate
           - 'isLimitedProxy' : True if the client use limited proxy certificate
           - 'group' (optional): Dirac group attached to the client
           - 'extraCredentials' (optional): Extra credentials if exists

          Before the handshake, dictionary is empty
        """
        return self.peerCredentials

    def setExtraCredentials(self, extraCredentials):
        """Add extra credentials to peerCredentials

        :param extraCredentials: group or tuple with DN and group
        :type extraCredentials: str or tuple
        """
        self.peerCredentials["extraCredentials"] = extraCredentials

    def serverMode(self):
        return self.bServerMode

    def getRemoteAddress(self):
        return self.remoteAddress

    def getLocalAddress(self):
        return self.oSocket.getsockname()

    def getSocket(self):
        return self.oSocket

    def _readReady(self):
        if not self.iReadTimeout:
            return True
        sel = selectors.DefaultSelector()
        sel.register(self.oSocket, selectors.EVENT_READ)
        if sel.select(timeout=self.iReadTimeout):
            return True
        return False

    def _read(self, bufSize=4096, skipReadyCheck=False):
        try:
            if skipReadyCheck or self._readReady():
                data = self.oSocket.recv(bufSize)
                if not data:
                    return S_ERROR("Connection closed by peer")
                else:
                    return S_OK(data)
            else:
                return S_ERROR("Connection seems stalled. Closing...")
        except Exception as e:
            return S_ERROR("Exception while reading from peer: %s" % str(e))

    def _write(self, buf):
        return S_OK(self.oSocket.send(buf))

    def sendData(self, uData, prefix=b""):
        self.__updateLastActionTimestamp()
        sCodedData = MixedEncode.encode(uData)
        if isinstance(sCodedData, six.text_type):
            sCodedData = sCodedData.encode()
        dataToSend = b"".join([prefix, str(len(sCodedData)).encode(), b":", sCodedData])
        for index in range(0, len(dataToSend), self.packetSize):
            bytesToSend = min(self.packetSize, len(dataToSend) - index)
            packSentBytes = 0
            while packSentBytes < bytesToSend:
                try:
                    result = self._write(dataToSend[index + packSentBytes : index + bytesToSend])
                    if not result["OK"]:
                        return result
                    sentBytes = result["Value"]
                except Exception as e:
                    return S_ERROR("Exception while sending data: %s" % e)
                if sentBytes == 0:
                    return S_ERROR("Connection closed by peer")
                packSentBytes += sentBytes
        del sCodedData
        sCodedData = None
        return S_OK()

    def receiveData(self, maxBufferSize=0, blockAfterKeepAlive=True, idleReceive=False):
        self.__updateLastActionTimestamp()
        if self.receivedMessages:
            return self.receivedMessages.pop(0)
        # Buffer size can't be less than 0
        maxBufferSize = max(maxBufferSize, 0)
        try:
            # Look either for message length of keep alive magic string
            iSeparatorPosition = self.byteStream.find(b":", 0, 10)
            keepAliveMagicLen = len(BaseTransport.keepAliveMagic)
            isKeepAlive = self.byteStream.find(BaseTransport.keepAliveMagic, 0, keepAliveMagicLen) == 0
            # While not found the message length or the ka, keep receiving
            while iSeparatorPosition == -1 and not isKeepAlive:
                retVal = self._read(16384)
                # If error return
                if not retVal["OK"]:
                    return retVal
                # If closed return error
                if not retVal["Value"]:
                    return S_ERROR("Peer closed connection")
                # New data!
                self.byteStream += retVal["Value"]
                # Look again for either message length of ka magic string
                iSeparatorPosition = self.byteStream.find(b":", 0, 10)
                isKeepAlive = self.byteStream.find(BaseTransport.keepAliveMagic, 0, keepAliveMagicLen) == 0
                # Over the limit?
                if maxBufferSize and len(self.byteStream) > maxBufferSize and iSeparatorPosition == -1:
                    return S_ERROR("Read limit exceeded (%s chars)" % maxBufferSize)
            # Keep alive magic!
            if isKeepAlive:
                gLogger.debug("Received keep alive header")
                # Remove the ka magic from the buffer and process the keep alive
                self.byteStream = self.byteStream[keepAliveMagicLen:]
                return self.__processKeepAlive(maxBufferSize, blockAfterKeepAlive)
            # From here it must be a real message!
            # Process the size and remove the msg length from the bytestream
            pkgSize = int(self.byteStream[:iSeparatorPosition])
            pkgData = self.byteStream[iSeparatorPosition + 1 :]
            readSize = len(pkgData)
            if readSize >= pkgSize:
                # If we already have all the data we need
                data = pkgData[:pkgSize]
                self.byteStream = pkgData[pkgSize:]
            else:
                # If we still need to read stuff
                pkgMem = BytesIO()
                pkgMem.write(pkgData)
                # Receive while there's still data to be received
                while readSize < pkgSize:
                    retVal = self._read(pkgSize - readSize, skipReadyCheck=True)
                    if not retVal["OK"]:
                        return retVal
                    if not retVal["Value"]:
                        return S_ERROR("Peer closed connection")
                    rcvData = retVal["Value"]
                    readSize += len(rcvData)
                    pkgMem.write(rcvData)
                    if maxBufferSize and readSize > maxBufferSize:
                        return S_ERROR("Read limit exceeded (%s chars)" % maxBufferSize)
                # Data is here! take it out from the bytestream, dencode and return
                if readSize == pkgSize:
                    data = pkgMem.getvalue()
                    self.byteStream = b""
                else:  # readSize > pkgSize:
                    pkgMem.seek(0, 0)
                    data = pkgMem.read(pkgSize)
                    self.byteStream = pkgMem.read()
            try:
                data = MixedEncode.decode(data)[0]
            except Exception as e:
                return S_ERROR("Could not decode received data: %s" % str(e))
            if idleReceive:
                self.receivedMessages.append(data)
                return S_OK()
            return data
        except Exception as e:
            gLogger.exception("Network error while receiving data")
            return S_ERROR("Network error while receiving data: %s" % str(e))

    def __processKeepAlive(self, maxBufferSize, blockAfterKeepAlive=True):
        gLogger.debug("Received Keep Alive")
        # Next message down the stream will be the ka data
        result = self.receiveData(maxBufferSize, blockAfterKeepAlive=False)
        if not result["OK"]:
            gLogger.debug("Error while receiving keep alive: %s" % result["Message"])
            return result
        # Is it a valid ka?
        kaData = result["Value"]
        for reqField in ("id", "kaping"):
            if reqField not in kaData:
                errMsg = "Invalid keep alive, missing %s" % reqField
                gLogger.debug(errMsg)
                return S_ERROR(errMsg)
        gLogger.debug("Received keep alive id %s" % kaData)
        # Need to check if it's one of the keep alives we sent or one started from the other side
        if kaData["kaping"]:
            # This is a keep alive PING. Let's send the PONG
            self.sendKeepAlive(responseId=kaData["id"])
        else:
            # If it's a pong then we flag that we don't need to wait for a pong
            self.waitingForKeepAlivePong = False
        # No blockAfterKeepAlive means return without further read
        if not blockAfterKeepAlive:
            result = S_OK()
            result["keepAlive"] = True
            return result
        # Let's listen for the next message downstream
        return self.receiveData(maxBufferSize, blockAfterKeepAlive)

    def sendKeepAlive(self, responseId=None, now=False):
        # If not responseId or not keepAliveLapse or not enough time has passed don't send keep alive
        if not responseId:
            if not self.__keepAliveLapse:
                return S_OK()
            if not now:
                now = time.time()
            if now - self.__lastActionTimestamp < self.__keepAliveLapse:
                return S_OK()

        self.__updateLastActionTimestamp()
        if responseId:
            self.waitingForKeepAlivePong = False
            kaData = S_OK({"id": responseId, "kaping": False})
        else:
            if self.waitingForKeepAlivePong:
                return S_OK()
            idK = self.keepAliveId + str(self.sentKeepAlives)
            self.sentKeepAlives += 1
            kaData = S_OK({"id": idK, "kaping": True})
            self.waitingForKeepAlivePong = True
        return self.sendData(kaData, prefix=BaseTransport.keepAliveMagic)

    def getFormattedCredentials(self):
        peerCreds = self.getConnectingCredentials()
        address = self.getRemoteAddress()
        if "username" in peerCreds:
            peerId = "[%s:%s]" % (peerCreds["group"], peerCreds["username"])
        else:
            peerId = ""
        if address[0].find(":") > -1:
            return "([%s]:%s)%s" % (address[0], address[1], peerId)
        return "(%s:%s)%s" % (address[0], address[1], peerId)

    def setSocketTimeout(self, timeout):
        """
        This method has to be overwritten, if we want to increase the socket timeout.
        """
        pass



""" PilotsLoggingDB class is a front-end to the Pilots Logging Database.
    This database keeps track of all the submitted grid pilot jobs.
    It also registers the mapping of the DIRAC jobs to the pilot
    agents.

    Available methods are:

    addPilotsLogging()
    getPilotsLogging()
    deletePilotsLoggin()

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import six
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters
from DIRAC.ResourceStatusSystem.Utilities import Utils

from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy import create_engine, Column, MetaData, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import SQLAlchemyError


TABLESLIST = ["PilotsLogging"]

metadata = MetaData()
Base = declarative_base()


#############################################################################
class PilotsLoggingDB(object):
    def __init__(self):

        self.log = gLogger.getSubLogger("PilotsLoggingDB")

        result = getDBParameters("WorkloadManagement/PilotsLoggingDB")
        if not result["OK"]:
            raise RuntimeError("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self.dbHost = dbParameters["Host"]
        self.dbPort = dbParameters["Port"]
        self.dbUser = dbParameters["User"]
        self.dbPass = dbParameters["Password"]
        self.dbName = dbParameters["DBName"]

        # These are the list of tables that will be created.
        # They can be extended in an extension module
        self.tablesList = getattr(Utils.voimport("DIRAC.WorkloadManagementSystem.DB.PilotsLoggingDB"), "TABLESLIST")

        self.__initializeConnection()
        resp = self.__initializeDB()
        if not resp["OK"]:
            raise Exception("Couldn't create tables: " + resp["Message"])

    ##########################################################################################

    def __initializeConnection(self):
        """
        This should be in a base class eventually
        """

        self.engine = create_engine(
            "mysql://%s:%s@%s:%s/%s" % (self.dbUser, self.dbPass, self.dbHost, self.dbPort, self.dbName),
            pool_recycle=3600,
            echo_pool=True,
            echo=self.log.getLevel() == "DEBUG",
        )
        self.sqlalchemySession = scoped_session(sessionmaker(bind=self.engine))
        self.inspector = Inspector.from_engine(self.engine)

    ##########################################################################################
    def __initializeDB(self):
        """
        Create the tables, if they are not there yet
        """

        tablesInDB = self.inspector.get_table_names()

        for table in self.tablesList:
            if table not in tablesInDB:
                found = False
                # is it in the extension? (fully or extended)
                for ext in gConfig.getValue("DIRAC/Extensions", []):
                    try:
                        getattr(__import__(ext + __name__, globals(), locals(), [table]), table).__table__.create(
                            self.engine
                        )  # pylint: disable=no-member
                        found = True
                        break
                    except (ImportError, AttributeError):
                        continue
                    # If not found in extensions, import it from DIRAC base.
                if not found:
                    getattr(__import__(__name__, globals(), locals(), [table]), table).__table__.create(
                        self.engine
                    )  # pylint: disable=no-member
            else:
                gLogger.debug("Table %s already exists" % table)

        return S_OK()

    ##########################################################################################
    def addPilotsLogging(self, pilotUUID, timestamp, source, phase, status, messageContent):
        """Add new pilot logging entry"""

        session = self.sqlalchemySession()
        logging = PilotsLogging(pilotUUID, timestamp, source, phase, status, messageContent)

        try:
            session.add(logging)
        except SQLAlchemyError as e:
            session.rollback()
            session.close()
            return S_ERROR(DErrno.ESQLA, "Failed to add PilotsLogging: " + str(e))

        try:
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            session.close()
            return S_ERROR(DErrno.ESQLA, "Failed to commit PilotsLogging: " + str(e))

        return S_OK()

    ##########################################################################################
    def getPilotsLogging(self, pilotUUID):
        """Get list of logging entries for pilot"""

        session = self.sqlalchemySession()

        pilotLogging = []
        for pl in (
            session.query(PilotsLogging)
            .filter(PilotsLogging.pilotUUID == pilotUUID)
            .order_by(PilotsLogging.timestamp)
            .all()
        ):
            entry = {}
            entry["pilotUUID"] = pl.pilotUUID
            entry["timestamp"] = pl.timestamp
            entry["source"] = pl.source
            entry["phase"] = pl.phase
            entry["status"] = pl.status
            entry["messageContent"] = pl.messageContent
            pilotLogging.append(entry)

        return S_OK(pilotLogging)

    ##########################################################################################
    def deletePilotsLogging(self, pilotUUID):
        """Delete all logging entries for pilot"""

        if isinstance(pilotUUID, six.string_types):
            pilotUUID = [
                pilotUUID,
            ]

        session = self.sqlalchemySession()

        session.query(PilotsLogging).filter(PilotsLogging.pilotUUID.in_(pilotUUID)).delete(synchronize_session="fetch")

        try:
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            session.close()
            return S_ERROR(DErrno.ESQLA, "Failed to commit: " + str(e))

        return S_OK()

    ##########################################################################################


class PilotsLogging(Base):
    """PilotsLogging table"""

    __tablename__ = "PilotsLogging"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    logID = Column("LogID", Integer, primary_key=True, autoincrement=True)
    pilotUUID = Column("pilotUUID", String(255), nullable=False)
    timestamp = Column("timestamp", String(255), nullable=False)
    source = Column("source", String(255), nullable=False)
    phase = Column("phase", String(255), nullable=False)
    status = Column("status", String(255), nullable=False)
    messageContent = Column("messageContent", String(255), nullable=False)

    def __init__(self, pilotUUID, timestamp, source, phase, status, messageContent):
        self.pilotUUID = pilotUUID
        self.timestamp = timestamp
        self.source = source
        self.phase = phase
        self.status = status
        self.messageContent = messageContent



"""
Command Line Parameters for creating the Replication transformations Script
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOMSVOForGroup


class Params(object):
    """Parameter Object"""

    def __init__(self):
        self.targetSE = []
        self.sourceSE = ""
        self.groupSize = 1
        self.groupName = None
        self.extraname = ""
        self.flavour = "Replication"
        self.plugin = "Broadcast"
        self.metaValues = []
        self.metaKey = None
        self.extraData = {}
        self.errorMessages = []
        self.enable = False

    def setMetaValues(self, values):
        if isinstance(values, list):
            self.metaValues = values
        else:
            self.metaValues = [val for val in values.split(",")]
        return S_OK()

    def setMetaKey(self, key):
        self.metaKey = key
        return S_OK()

    def setMetadata(self, metadata):
        for pair in metadata.split(","):
            splitPair = pair.strip().split(":")
            if len(splitPair) == 2:
                self.extraData[splitPair[0]] = splitPair[1].strip()
        return S_OK()

    def setSourceSE(self, sourceSE):
        self.sourceSE = [sSE.strip() for sSE in sourceSE.split(",")]
        return S_OK()

    def setTransFlavour(self, flavour):
        self.flavour = flavour
        return S_OK()

    def setTargetSE(self, targetSE):
        self.targetSE = [tSE.strip() for tSE in targetSE.split(",")]
        return S_OK()

    def setExtraname(self, extraname):
        self.extraname = extraname
        return S_OK()

    def setGroupSize(self, size):
        try:
            self.groupSize = int(size)
        except ValueError:
            return S_ERROR("Expected integer for groupsize")
        return S_OK()

    def setGroupName(self, name):
        self.groupName = name
        return S_OK()

    def setPlugin(self, plugin):
        self.plugin = plugin
        return S_OK()

    def setEnable(self, _):
        self.enable = True
        return S_OK()

    def registerSwitches(self, script):
        """register command line arguments

        :param script: Dirac.Core.Base Script Class
        :type script: DIRAC.Core.Base.Script
        """

        script.registerSwitch("G:", "GroupSize=", "Number of Files per transformation task", self.setGroupSize)
        script.registerSwitch("R:", "GroupName=", "TransformationGroup Name", self.setGroupName)
        script.registerSwitch("S:", "SourceSEs=", "SourceSE(s) to use, comma separated list", self.setSourceSE)
        script.registerSwitch("N:", "Extraname=", "String to append to transformation name", self.setExtraname)
        script.registerSwitch("P:", "Plugin=", "Plugin to use for transformation", self.setPlugin)
        script.registerSwitch("T:", "Flavour=", "Flavour to create: Replication or Moving", self.setTransFlavour)
        script.registerSwitch("K:", "MetaKey=", "Meta Key to use: TransformationID", self.setMetaKey)
        script.registerSwitch("M:", "MetaData=", "MetaData to use Key/Value Pairs: 'DataType:REC,'", self.setMetadata)
        script.registerSwitch("x", "Enable", "Enable the transformation creation, otherwise dry-run", self.setEnable)

        useMessage = []
        useMessage.append("Create one replication transformation for each MetaValue given")
        useMessage.append("Is running in dry-run mode, unless enabled with -x")
        useMessage.append("MetaValue and TargetSEs can be comma separated lists")
        useMessage.append("Usage:")
        useMessage.append(
            "  %s <MetaValue1[,val2,val3]> <TargetSEs> [-G<Files>] [-S<SourceSEs>]"
            "[-N<ExtraName>] [-T<Type>] [-M<Key>] [-K...] -x" % script.scriptName
        )
        script.setUsageMessage("\n".join(useMessage))

    def checkSettings(self, script, checkArguments=True):
        """check if all required parameters are set, print error message and return S_ERROR if not

        :param script: The script object
        :type script: DIRAC.Core.Base.Script
        :param bool checkArguments: if false do not check for the correct number of arguments, should only be
            changed if using derived class
        """

        if checkArguments:
            args = script.getPositionalArgs()
            if len(args) == 2:
                self.setMetaValues(args[0])
                self.setTargetSE(args[1])
            else:
                self.errorMessages.append("ERROR: Wrong number of arguments")

        self._checkProxy()

        # get default metadata key:
        from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

        if self.metaKey is None:
            self.metaKey = Operations().getValue("Transformations/TransfIDMeta", "TransformationID")

        if not self.errorMessages:
            return S_OK()
        gLogger.error("\n".join(self.errorMessages))
        script.showHelp()
        return S_ERROR()

    def _checkProxy(self):
        """checks if the proxy has the ProductionManagement property and belongs to a VO"""
        proxyInfo = getProxyInfo()
        if not proxyInfo["OK"]:
            self.errorMessages.append("ERROR: No Proxy present")
            return False
        proxyValues = proxyInfo.get("Value", {})
        group = proxyValues.get("group", "")
        vomsvo = getVOMSVOForGroup(group)
        if not vomsvo:
            self.errorMessages.append("ERROR: ProxyGroup not associated to VOMS VO, get a different proxy")
            return False

        groupProperties = proxyValues.get("groupProperties", [])

        if groupProperties:
            if "ProductionManagement" not in groupProperties:
                self.errorMessages.append(
                    "ERROR: Not allowed to create production, you need a ProductionManagement proxy."
                )
                return False
        else:
            self.errorMessages.append("ERROR: Could not determine Proxy properties, you do not have the right proxy.")
            return False
        return True



"""
PropagationCommand module
This command gets all the elements that exist under a given site and applies the following logic:
if even one element is 'Active' for the given site then it marks the site as 'Active', if all elements
are set in a different status (like banned or error) then it marks the site as 'Banned'
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Command.Command import Command
from DIRAC.ResourceStatusSystem.Utilities import CSHelpers


class PropagationCommand(Command):
    def __init__(self, args=None, clients=None):

        self.rssClient = ResourceStatusClient()
        super(PropagationCommand, self).__init__(args, clients)

    def doNew(self, masterParams=None):
        return S_OK()

    def doCache(self):

        if not self.args["name"]:
            return S_ERROR("site was not found in args")

        site = self.args["name"]

        elements = CSHelpers.getSiteElements(site)

        statusList = []

        if elements["OK"]:
            for element in elements["Value"]:
                status = self.rssClient.selectStatusElement("Resource", "Status", element, meta={"columns": ["Status"]})
                if not status["OK"]:
                    return status

                if status["Value"]:
                    statusList.append(status["Value"][0][0])
                else:  # forcing in the case the resource has no status (yet)
                    statusList.append("Active")

            if "Active" in statusList:
                return S_OK({"Status": "Active", "Reason": "An element that belongs to the site is Active"})

            if "Degraded" in statusList:
                return S_OK({"Status": "Degraded", "Reason": "An element that belongs to the site is Degraded"})

        return S_OK({"Status": "Banned", "Reason": "There is no Active element in the site"})

    def doMaster(self):
        return S_OK()



# Creating 3 processes, for tests,
# that consume some CPU and that have children

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
from multiprocessing import Pool, Process, current_process


def f(n):
    time.sleep(n)


def do_sum(li):
    sum(li)
    proc_name = current_process().name
    print(proc_name)


my_list = list(range(150000000))

pool = Pool(3)
pool.map(do_sum, [my_list[:50000000], my_list[50000000:100000000], my_list[100000000:]])
pool.close()
pool.join()

# Generate system CPU usage
for i in range(1000):
    p = Process(target=f, args=(0.001,))
    p.daemon = False
    p.start()
    p.join()



"""
DIRAC Times module
Support for basic Date and Time operations
based on system datetime module.

It provides common interface to UTC timestamps,
converter to string types and back.

The following datetime classes are used in the returned objects:

  - dateTime  = datetime.datetime
  - date      = datetime.date
  - time      = datetime.timedelta

Useful timedelta constant are also provided to
define time intervals.

Notice: datetime.timedelta objects allow multiplication and division by interger
but not by float. Thus:

  - DIRAC.Times.second * 1.5             is not allowed
  - DIRAC.Times.second * 3 / 2           is allowed

An timeInterval class provides a method to check
if a give datetime is in the defined interval.

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import time as nativetime
import datetime
import six
import sys

__RCSID__ = "$Id$"


# Some useful constants for time operations
microsecond = datetime.timedelta(microseconds=1)
second = datetime.timedelta(seconds=1)
minute = datetime.timedelta(minutes=1)
hour = datetime.timedelta(hours=1)
day = datetime.timedelta(days=1)
week = datetime.timedelta(days=7)

dt = datetime.datetime(2000, 1, 1)


def timeThis(method):
    """Function to be used as a decorator for timing other functions/methods"""

    def timed(*args, **kw):
        """What actually times"""
        ts = nativetime.time()
        result = method(*args, **kw)
        if sys.stdout.isatty():
            return result
        te = nativetime.time()

        pre = dt.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC ")

        try:
            pre += args[0].log.getName() + "/" + args[0].log.getSubName() + "   TIME: " + args[0].transString
        except AttributeError:
            try:
                pre += args[0].log.getName() + "    TIME: " + args[0].transString
            except AttributeError:
                try:
                    pre += args[0].log.getName() + "/" + args[0].log.getSubName() + "   TIME: "
                except AttributeError:
                    pre += "TIME: "
        except IndexError:
            pre += "TIME: "

        argsLen = ""
        if args:
            try:
                if isinstance(args[1], (list, dict)):
                    argsLen = "arguments len: %d" % len(args[1])
            except IndexError:
                if kw:
                    try:
                        if isinstance(list(list(kw.items())[0])[1], (list, dict)):
                            argsLen = "arguments len: %d" % len(list(list(kw.items())[0])[1])
                    except IndexError:
                        argsLen = ""

        print("%s Exec time ===> function %r %s -> %2.2f sec" % (pre, method.__name__, argsLen, te - ts))
        return result

    return timed


def dateTime():
    """
    Return current UTC datetime, as datetime.datetime object
    """
    return dt.utcnow()


def date(myDateTime=None):
    """
    Return current UTC date, as datetime.date object
    if a _dateTimeType is pass as argument its associated date is returned
    """
    if isinstance(myDateTime, _dateTimeType):
        return myDateTime.date()
    return dateTime().date()


def time(myDateTime=None):
    """
    Return current UTC time, as datetime.time object
    if a _dateTimeType is pass as argument its associated time is returned
    """
    if not isinstance(myDateTime, _dateTimeType):
        myDateTime = dateTime()
    return myDateTime - datetime.datetime(myDateTime.year, myDateTime.month, myDateTime.day)


def toEpoch(dateTimeObject=None):
    """
    Get seconds since epoch
    """
    if not dateTimeObject:
        dateTimeObject = dateTime()
    return nativetime.mktime(dateTimeObject.timetuple())


def fromEpoch(epoch):
    """
    Get datetime object from epoch
    """
    return dt.fromtimestamp(epoch)


def to2K(dateTimeObject=None):
    """
    Get seconds, with microsecond precission, since 2K
    """
    if not dateTimeObject:
        dateTimeObject = dateTime()
    delta = dateTimeObject - dt
    return delta.days * 86400 + delta.seconds + delta.microseconds / 1000000.0


def from2K(seconds2K=None):
    """
    Get date from seconds since 2K
    """
    if not seconds2K:
        seconds2K = to2K(dt)
    return dt + int(seconds2K) * second + int(seconds2K % 1 * 1000000) * microsecond


def toString(myDate=None):
    """
    Convert to String
    if argument type is neither _dateTimeType, _dateType, nor _timeType
    the current dateTime converted to String is returned instead

    Notice: datetime.timedelta are converted to strings using the format:
      [day] days [hour]:[min]:[sec]:[microsec]
      where hour, min, sec, microsec are always positive integers,
      and day carries the sign.
    To keep internal consistency we are using:
      [hour]:[min]:[sec]:[microsec]
      where min, sec, microsec are alwys positive intergers and hour carries the
      sign.
    """
    if isinstance(myDate, _dateTimeType):
        return str(myDate)

    elif isinstance(myDate, _dateType):
        return str(myDate)

    elif isinstance(myDate, _timeType):
        return "%02d:%02d:%02d.%06d" % (
            myDate.days * 24 + myDate.seconds / 3600,
            myDate.seconds % 3600 / 60,
            myDate.seconds % 60,
            myDate.microseconds,
        )
    else:
        return toString(dateTime())


def fromString(myDate=None):
    """
    Convert date/time/datetime String back to appropriated objects

    The format of the string it is assume to be that returned by toString method.
    See notice on toString method
    On Error, return None
    """
    if isinstance(myDate, six.string_types):
        if myDate.find(" ") > 0:
            dateTimeTuple = myDate.split(" ")
            dateTuple = dateTimeTuple[0].split("-")
            try:
                return datetime.datetime(year=dateTuple[0], month=dateTuple[1], day=dateTuple[2]) + fromString(
                    dateTimeTuple[1]
                )
                # return dt.combine( fromString( dateTimeTuple[0] ),
                #                                   fromString( dateTimeTuple[1] ) )
            except Exception:
                try:
                    return datetime.datetime(
                        year=int(dateTuple[0]), month=int(dateTuple[1]), day=int(dateTuple[2])
                    ) + fromString(dateTimeTuple[1])
                except ValueError:
                    return None
                # return dt.combine( fromString( dateTimeTuple[0] ),
                #                                   fromString( dateTimeTuple[1] ) )
        elif myDate.find(":") > 0:
            timeTuple = myDate.replace(".", ":").split(":")
            try:
                if len(timeTuple) == 4:
                    return datetime.timedelta(
                        hours=int(timeTuple[0]),
                        minutes=int(timeTuple[1]),
                        seconds=int(timeTuple[2]),
                        microseconds=int(timeTuple[3]),
                    )
                elif len(timeTuple) == 3:
                    try:
                        return datetime.timedelta(
                            hours=int(timeTuple[0]),
                            minutes=int(timeTuple[1]),
                            seconds=int(timeTuple[2]),
                            microseconds=0,
                        )
                    except ValueError:
                        return None
                else:
                    return None
            except Exception:
                return None
        elif myDate.find("-") > 0:
            dateTuple = myDate.split("-")
            try:
                return datetime.date(int(dateTuple[0]), int(dateTuple[1]), int(dateTuple[2]))
            except Exception:
                return None

    return None


class timeInterval(object):
    """
    Simple class to define a timeInterval object able to check if a given
    dateTime is inside
    """

    def __init__(self, initialDateTime, intervalTimeDelta):
        """
        Initialization method, it requires the initial dateTime and the
        timedelta that define the limits.
        The upper limit is not included thus it is [begin,end)
        If not properly initialized an error flag is set, and subsequent calls
        to any method will return None
        """
        if not isinstance(initialDateTime, _dateTimeType) or not isinstance(intervalTimeDelta, _timeType):
            self.__error = True
            return None
        self.__error = False
        if intervalTimeDelta.days < 0:
            self.__startDateTime = initialDateTime + intervalTimeDelta
            self.__endDateTime = initialDateTime
        else:
            self.__startDateTime = initialDateTime
            self.__endDateTime = initialDateTime + intervalTimeDelta

    def includes(self, myDateTime):
        """ """
        if self.__error:
            return None
        if not isinstance(myDateTime, _dateTimeType):
            return None
        if myDateTime < self.__startDateTime:
            return False
        if myDateTime >= self.__endDateTime:
            return False
        return True


def queryTime(f):
    """Decorator to measure the function call time"""

    def measureQueryTime(*args, **kwargs):
        start = nativetime.time()
        result = f(*args, **kwargs)
        if result["OK"] and "QueryTime" not in result:
            result["QueryTime"] = nativetime.time() - start
        return result

    return measureQueryTime


_dateTimeType = type(dateTime())
_dateType = type(date())
_timeType = type(time())

_allTimeTypes = (_dateTimeType, _timeType)
_allDateTypes = (_dateTimeType, _dateType)
_allTypes = (_dateTimeType, _dateType, _timeType)



"""
  The Job Path executor determines the chain of Optimizing Agents that must
  work on the job prior to the scheduling decision.

  Initially this takes jobs in the received state and starts the jobs on the
  optimizer chain.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import List
from DIRAC.WorkloadManagementSystem.Executor.Base.OptimizerExecutor import OptimizerExecutor


class JobPath(OptimizerExecutor):
    """
    The specific Optimizer must provide the following methods:
    - optimizeJob() - the main method called for each job
    and it can provide:
    - initializeOptimizer() before each execution cycle
    """

    @classmethod
    def initializeOptimizer(cls):
        cls.__voPlugins = {}
        return S_OK()

    def __setOptimizerChain(self, jobState, opChain):
        if not isinstance(opChain, six.string_types):
            opChain = ",".join(opChain)
        return jobState.setOptParameter("OptimizerChain", opChain)

    def __executeVOPlugin(self, voPlugin, jobState):
        if voPlugin not in self.__voPlugins:
            modName = List.fromChar(voPlugin, ".")[-1]
            try:
                module = __import__(voPlugin, globals(), locals(), [modName])
            except ImportError:
                self.jobLog.exception("Could not import VO plugin", voPlugin)
                return S_ERROR("Could not import VO plugin")

            try:
                self.__voPlugins[voPlugin] = getattr(module, modName)
            except AttributeError as excp:
                return S_ERROR("Could not get plugin %s from module %s: %s" % (modName, voPlugin, str(excp)))

        argsDict = {"JobID": jobState.jid, "JobState": jobState, "ConfigPath": self.ex_getProperty("section")}
        try:
            modInstance = self.__voPlugins[voPlugin](argsDict)
            result = modInstance.execute()
        except Exception:
            self.jobLog.exception("Excp while executing", voPlugin)
            return S_ERROR("Could not execute VO plugin")

        if not result["OK"]:
            return result
        extraPath = result["Value"]
        if isinstance(extraPath, six.string_types):
            extraPath = List.fromChar(result["Value"])
        return S_OK(extraPath)

    def optimizeJob(self, jid, jobState):
        result = jobState.getManifest()
        if not result["OK"]:
            self.jobLog.error("Failed to get job manifest", result["Message"])
            return result
        jobManifest = result["Value"]
        opChain = jobManifest.getOption("JobPath", [])
        if opChain:
            self.jobLog.info("Job defines its own optimizer chain", opChain)
            result = self.__setOptimizerChain(jobState, opChain)
            if not result["OK"]:
                self.jobLog.error("Failed to set optimizer chain", result["Message"])
                return result
            return self.setNextOptimizer(jobState)
        # Construct path
        opPath = self.ex_getOption("BasePath", ["JobPath", "JobSanity"])
        voPlugin = self.ex_getOption("VOPlugin", "")
        # Specific VO path
        if voPlugin:
            result = self.__executeVOPlugin(voPlugin, jobState)
            if not result["OK"]:
                return result
            extraPath = result["Value"]
            if extraPath:
                opPath.extend(extraPath)
                self.jobLog.verbose("Adding extra VO specific optimizers to path", extraPath)
        else:
            # Generic path: Should only rely on an input data setting in absence of VO plugin
            self.jobLog.verbose("No VO specific plugin module specified")
            result = jobState.getInputData()
            if not result["OK"]:
                self.jobLog.error("Failed to get input data", result["Message"])
                return result
            if result["Value"]:
                # if the returned tuple is not empty it will evaluate true
                self.jobLog.info("Input data requirement found")
                opPath.extend(self.ex_getOption("InputData", ["InputData"]))
            else:
                self.jobLog.info("No input data requirement")
        # End of path
        opPath.extend(self.ex_getOption("EndPath", ["JobScheduling"]))
        uPath = []
        for opN in opPath:
            if opN not in uPath:
                uPath.append(opN)
        opPath = uPath
        self.jobLog.info("Constructed path is", "%s" % "->".join(opPath))
        result = self.__setOptimizerChain(jobState, opPath)
        if not result["OK"]:
            self.jobLog.error("Failed to set optimizer chain", result["Message"])
            return result
        return self.setNextOptimizer(jobState)



""" The FileCatalogClient is a class representing the client of the DIRAC File Catalog
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Tornado.Client.ClientSelector import TransferClientSelector as TransferClient

from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOMSAttributeForGroup, getDNForUsername
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase

__RCSID__ = "$Id$"


class FileCatalogClient(FileCatalogClientBase):
    """Client code to the DIRAC File Catalogue"""

    # The list of methods below is defining the client interface
    READ_METHODS = FileCatalogClientBase.READ_METHODS + [
        "isFile",
        "getFileMetadata",
        "getReplicas",
        "getReplicaStatus",
        "getFileSize",
        "isDirectory",
        "getDirectoryReplicas",
        "listDirectory",
        "getDirectoryMetadata",
        "getDirectorySize",
        "getDirectoryContents",
        "getLFNForPFN",
        "getLFNForGUID",
        "findFilesByMetadata",
        "getMetadataFields",
        "findDirectoriesByMetadata",
        "getReplicasByMetadata",
        "findFilesByMetadataDetailed",
        "findFilesByMetadataWeb",
        "getCompatibleMetadata",
        "getMetadataSet",
        "getDatasets",
        "getFileDescendents",
        "getFileAncestors",
        "getDirectoryUserMetadata",
        "getFileUserMetadata",
        "checkDataset",
        "getDatasetParameters",
        "getDatasetFiles",
        "getDatasetAnnotation",
    ]

    WRITE_METHODS = [
        "createLink",
        "removeLink",
        "addFile",
        "addFileAncestors",
        "setFileStatus",
        "addReplica",
        "removeReplica",
        "removeFile",
        "setReplicaStatus",
        "setReplicaHost",
        "setReplicaProblematic",
        "createDirectory",
        "setDirectoryStatus",
        "removeDirectory",
        "changePathMode",
        "changePathOwner",
        "changePathGroup",
        "addMetadataField",
        "deleteMetadataField",
        "setMetadata",
        "setMetadataBulk",
        "removeMetadata",
        "addMetadataSet",
        "addDataset",
        "addDatasetAnnotation",
        "removeDataset",
        "updateDataset",
        "freezeDataset",
        "releaseDataset",
        "addUser",
        "deleteUser",
        "addGroup",
        "deleteGroup",
        "repairCatalog",
        "rebuildDirectoryUsage",
    ]

    NO_LFN_METHODS = [
        "findFilesByMetadata",
        "addMetadataField",
        "deleteMetadataField",
        "getMetadataFields",
        "setMetadata",
        "setMetadataBulk",
        "removeMetadata",
        "getDirectoryUserMetadata",
        "findDirectoriesByMetadata",
        "getReplicasByMetadata",
        "findFilesByMetadataDetailed",
        "findFilesByMetadataWeb",
        "getCompatibleMetadata",
        "addMetadataSet",
        "getMetadataSet",
        "getFileUserMetadata",
        "getLFNForGUID",
        "addUser",
        "deleteUser",
        "addGroup",
        "deleteGroup",
        "repairCatalog",
        "rebuildDirectoryUsage",
    ]

    ADMIN_METHODS = [
        "addUser",
        "deleteUser",
        "addGroup",
        "deleteGroup",
        "getUsers",
        "getGroups",
        "getCatalogCounters",
        "repairCatalog",
        "rebuildDirectoryUsage",
    ]

    def __init__(self, url=None, **kwargs):
        """Constructor function."""
        self.serverURL = "DataManagement/FileCatalog" if not url else url
        super(FileCatalogClient, self).__init__(self.serverURL, **kwargs)

    @checkCatalogArguments
    def getReplicas(self, lfns, allStatus=False, timeout=120):
        """Get the replicas of the given files"""
        rpcClient = self._getRPC(timeout=timeout)
        result = rpcClient.getReplicas(lfns, allStatus)

        if not result["OK"]:
            return result

        # If there is no PFN returned, just set the LFN instead
        lfnDict = result["Value"]
        for lfn in lfnDict["Successful"]:
            for se in lfnDict["Successful"][lfn]:
                if not lfnDict["Successful"][lfn][se]:
                    lfnDict["Successful"][lfn][se] = lfn

        return S_OK(lfnDict)

    @checkCatalogArguments
    def setReplicaProblematic(self, lfns, revert=False):
        """
        Set replicas to problematic.

        :param lfn lfns: has to be formated this way :
                    { lfn : { se1 : pfn1, se2 : pfn2, ...}, ...}
        :param revert: If True, remove the problematic flag

        :return: { successful : { lfn : [ ses ] } : failed : { lfn : { se : msg } } }
        """

        # This method does a batch treatment because the setReplicaStatus can only take one replica per lfn at once
        #
        # Illustration :
        #
        # lfns {'L2': {'S1': 'P3'}, 'L3': {'S3': 'P5', 'S2': 'P4', 'S4': 'P6'}, 'L1': {'S2': 'P2', 'S1': 'P1'}}
        #
        # loop1: lfnSEs {'L2': ['S1'], 'L3': ['S3', 'S2', 'S4'], 'L1': ['S2', 'S1']}
        # loop1 : batch {'L2': {'Status': 'P', 'SE': 'S1', 'PFN': 'P3'},
        #                'L3': {'Status': 'P', 'SE': 'S4', 'PFN': 'P6'},
        #                'L1': {'Status': 'P', 'SE': 'S1', 'PFN': 'P1'}}
        #
        # loop2: lfnSEs {'L2': [], 'L3': ['S3', 'S2'], 'L1': ['S2']}
        # loop2 : batch {'L3': {'Status': 'P', 'SE': 'S2', 'PFN': 'P4'}, 'L1': {'Status': 'P', 'SE': 'S2', 'PFN': 'P2'}}
        #
        # loop3: lfnSEs {'L3': ['S3'], 'L1': []}
        # loop3 : batch {'L3': {'Status': 'P', 'SE': 'S3', 'PFN': 'P5'}}
        #
        # loop4: lfnSEs {'L3': []}
        # loop4 : batch {}

        successful = {}
        failed = {}

        status = "AprioriGood" if revert else "Trash"

        # { lfn : [ se1, se2, ...], ...}
        lfnsSEs = dict((lfn, [se for se in lfns[lfn]]) for lfn in lfns)

        while lfnsSEs:

            # { lfn : { 'SE' : se1, 'PFN' : pfn1, 'Status' : status }, ... }
            batch = {}

            for lfn in list(lfnsSEs):
                # If there are still some Replicas (SE) for the given LFN, we put it in the next batch
                # else we remove the entry from the lfnsSEs dict
                if lfnsSEs[lfn]:
                    se = lfnsSEs[lfn].pop()
                    batch[lfn] = {"SE": se, "PFN": lfns[lfn][se], "Status": status}
                else:
                    del lfnsSEs[lfn]

            # Happens when there is nothing to treat anymore
            if not batch:
                break

            res = self.setReplicaStatus(batch)
            if not res["OK"]:
                for lfn in batch:
                    failed.setdefault(lfn, {})[batch[lfn]["SE"]] = res["Message"]
                continue

            for lfn in res["Value"]["Failed"]:
                failed.setdefault(lfn, {})[batch[lfn]["SE"]] = res["Value"]["Failed"][lfn]

            for lfn in res["Value"]["Successful"]:
                successful.setdefault(lfn, []).append(batch[lfn]["SE"])

        return S_OK({"Successful": successful, "Failed": failed})

    @checkCatalogArguments
    def listDirectory(self, lfn, verbose=False, timeout=120):
        """List the given directory's contents"""
        rpcClient = self._getRPC(timeout=timeout)
        result = rpcClient.listDirectory(lfn, verbose)
        if not result["OK"]:
            return result
        # Force returned directory entries to be LFNs
        for entryType in ["Files", "SubDirs", "Links"]:
            for path in result["Value"]["Successful"]:
                entryDict = result["Value"]["Successful"][path][entryType]
                for fname in list(entryDict):
                    detailsDict = entryDict.pop(fname)
                    lfn = os.path.join(path, os.path.basename(fname))
                    entryDict[lfn] = detailsDict
        return result

    @checkCatalogArguments
    def getDirectoryMetadata(self, lfns, timeout=120):
        """Get standard directory metadata"""
        rpcClient = self._getRPC(timeout=timeout)
        result = rpcClient.getDirectoryMetadata(lfns)
        if not result["OK"]:
            return result
        # Add some useful fields
        for path in result["Value"]["Successful"]:
            owner = result["Value"]["Successful"][path]["Owner"]
            group = result["Value"]["Successful"][path]["OwnerGroup"]
            res = getDNForUsername(owner)
            if res["OK"]:
                result["Value"]["Successful"][path]["OwnerDN"] = res["Value"][0]
            else:
                result["Value"]["Successful"][path]["OwnerDN"] = ""
            result["Value"]["Successful"][path]["OwnerRole"] = getVOMSAttributeForGroup(group)
        return result

    @checkCatalogArguments
    def removeDirectory(self, lfn, recursive=False, timeout=120):
        """Remove the directory from the File Catalog. The recursive keyword is for the ineterface."""
        rpcClient = self._getRPC(timeout=timeout)
        return rpcClient.removeDirectory(lfn)

    @checkCatalogArguments
    def getDirectoryReplicas(self, lfns, allStatus=False, timeout=120):
        """Find all the given directories' replicas"""
        rpcClient = self._getRPC(timeout=timeout)
        result = rpcClient.getDirectoryReplicas(lfns, allStatus)

        if not result["OK"]:
            return result

        # Replace just the filename with the full LFN
        for path in result["Value"]["Successful"]:
            pathDict = result["Value"]["Successful"][path]
            for fname in list(pathDict):
                # Remove the file name from the directory dict
                detailsDict = pathDict.pop(fname)
                # Build the lfn
                lfn = "%s/%s" % (path, os.path.basename(fname))
                # Add the LFN as value for each SE which does not have a PFN
                for se in detailsDict:
                    if not detailsDict[se]:
                        detailsDict[se] = lfn
                # Add the lfn entry to the directory dict
                pathDict[lfn] = detailsDict

        return result

    def findFilesByMetadata(self, metaDict, path="/", timeout=120):
        """Find files given the meta data query and the path"""
        rpcClient = self._getRPC(timeout=timeout)
        result = rpcClient.findFilesByMetadata(metaDict, path)
        if not result["OK"]:
            return result
        if isinstance(result["Value"], list):
            return result
        elif isinstance(result["Value"], dict):
            # Process into the lfn list
            fileList = []
            for dir_, fList in result["Value"].items():
                for fi in fList:
                    fileList.append(dir_ + "/" + fi)
            result["Value"] = fileList
            return result
        else:
            return S_ERROR("Illegal return value type %s" % type(result["Value"]))

    def getFileUserMetadata(self, path, timeout=120):
        """Get the meta data attached to a file, but also to
        the its corresponding directory
        """
        directory = "/".join(path.split("/")[:-1])
        rpcClient = self._getRPC(timeout=timeout)
        result = rpcClient.getFileUserMetadata(path)
        if not result["OK"]:
            return result
        fmeta = result["Value"]
        result = rpcClient.getDirectoryUserMetadata(directory)
        if not result["OK"]:
            return result
        fmeta.update(result["Value"])

        return S_OK(fmeta)

    ########################################################################
    # Path operations (not updated)
    #

    @checkCatalogArguments
    def changePathOwner(self, lfns, recursive=False, timeout=120):
        """Get replica info for the given list of LFNs"""
        return self._getRPC(timeout=timeout).changePathOwner(lfns, recursive)

    @checkCatalogArguments
    def changePathGroup(self, lfns, recursive=False, timeout=120):
        """Get replica info for the given list of LFNs"""
        return self._getRPC(timeout=timeout).changePathGroup(lfns, recursive)

    @checkCatalogArguments
    def changePathMode(self, lfns, recursive=False, timeout=120):
        """Get replica info for the given list of LFNs"""
        return self._getRPC(timeout=timeout).changePathMode(lfns, recursive)

    ########################################################################
    # ACL Operations
    #

    @checkCatalogArguments
    def getPathPermissions(self, lfns, timeout=120):
        """Determine the ACL information for a supplied path"""
        return self._getRPC(timeout=timeout).getPathPermissions(lfns)

    @checkCatalogArguments
    def hasAccess(self, paths, opType, timeout=120):
        """Determine if the given op can be performed on the paths
        The OpType is all the operations exported
        """
        return self._getRPC(timeout=timeout).hasAccess(paths, opType)

    ###################################################################
    #
    #  User/Group write operations
    #

    def addUser(self, userName, timeout=120):
        """Add a new user to the File Catalog"""
        return self._getRPC(timeout=timeout).addUser(userName)

    def deleteUser(self, userName, timeout=120):
        """Delete user from the File Catalog"""
        return self._getRPC(timeout=timeout).deleteUser(userName)

    def addGroup(self, groupName, timeout=120):
        """Add a new group to the File Catalog"""
        return self._getRPC(timeout=timeout).addGroup(groupName)

    def deleteGroup(self, groupName, timeout=120):
        """Delete group from the File Catalog"""
        return self._getRPC(timeout=timeout).deleteGroup(groupName)

    ###################################################################
    #
    #  User/Group read operations
    #

    def getUsers(self, timeout=120):
        """Get all the users defined in the File Catalog"""
        return self._getRPC(timeout=timeout).getUsers()

    def getGroups(self, timeout=120):
        """Get all the groups defined in the File Catalog"""
        return self._getRPC(timeout=timeout).getGroups()

    ########################################################################
    #
    # Path read operations
    #

    @checkCatalogArguments
    def exists(self, lfns, timeout=120):
        """Check whether the supplied paths exists"""
        return self._getRPC(timeout=timeout).exists(lfns)

    ########################################################################
    #
    # File write operations
    #

    @checkCatalogArguments
    def addFile(self, lfns, timeout=120):
        """Register supplied files"""

        return self._getRPC(timeout=timeout).addFile(lfns)

    @checkCatalogArguments
    def removeFile(self, lfns, timeout=120):
        """Remove the supplied lfns"""
        return self._getRPC(timeout=timeout).removeFile(lfns)

    @checkCatalogArguments
    def setFileStatus(self, lfns, timeout=120):
        """Remove the supplied lfns"""
        return self._getRPC(timeout=timeout).setFileStatus(lfns)

    @checkCatalogArguments
    def addReplica(self, lfns, timeout=120):
        """Register supplied replicas"""
        return self._getRPC(timeout=timeout).addReplica(lfns)

    @checkCatalogArguments
    def removeReplica(self, lfns, timeout=120):
        """Remove the supplied replicas"""
        return self._getRPC(timeout=timeout).removeReplica(lfns)

    @checkCatalogArguments
    def setReplicaStatus(self, lfns, timeout=120):
        """Set the status for the supplied replicas"""
        return self._getRPC(timeout=timeout).setReplicaStatus(lfns)

    @checkCatalogArguments
    def setReplicaHost(self, lfns, timeout=120):
        """Change the registered SE for the supplied replicas"""
        return self._getRPC(timeout=timeout).setReplicaHost(lfns)

    @checkCatalogArguments
    def addFileAncestors(self, lfns, timeout=120):
        """Add file ancestor information for the given dict of LFNs.

        :param dict lfns: {lfn1: {'Ancestor': [ancestorLFNs]}, lfn2: {'Ancestors': ...}}
        """
        return self._getRPC(timeout=timeout).addFileAncestors(lfns)

    ########################################################################
    #
    # File read operations
    #

    @checkCatalogArguments
    def isFile(self, lfns, timeout=120):
        """Check whether the supplied lfns are files"""
        return self._getRPC(timeout=timeout).isFile(lfns)

    @checkCatalogArguments
    def getFileSize(self, lfns, timeout=120):
        """Get the size associated to supplied lfns"""
        return self._getRPC(timeout=timeout).getFileSize(lfns)

    @checkCatalogArguments
    def getFileMetadata(self, lfns, timeout=120):
        """Get the metadata associated to supplied lfns"""
        return self._getRPC(timeout=timeout).getFileMetadata(lfns)

    @checkCatalogArguments
    def getReplicaStatus(self, lfns, timeout=120):
        """Get the status for the supplied replicas"""
        return self._getRPC(timeout=timeout).getReplicaStatus(lfns)

    @checkCatalogArguments
    def getFileAncestors(self, lfns, depths, timeout=120):
        """Get the status for the supplied replicas"""
        return self._getRPC(timeout=timeout).getFileAncestors(lfns, depths)

    @checkCatalogArguments
    def getFileDescendents(self, lfns, depths, timeout=120):
        """Get the status for the supplied replicas"""
        return self._getRPC(timeout=timeout).getFileDescendents(lfns, depths)

    def getLFNForGUID(self, guids, timeout=120):
        """Get the matching lfns for given guids"""
        return self._getRPC(timeout=timeout).getLFNForGUID(guids)

    ########################################################################
    #
    # Directory write operations
    #

    @checkCatalogArguments
    def createDirectory(self, lfns, timeout=120):
        """Create the supplied directories"""
        return self._getRPC(timeout=timeout).createDirectory(lfns)

    ########################################################################
    #
    # Directory read operations
    #

    @checkCatalogArguments
    def isDirectory(self, lfns, timeout=120):
        """Determine whether supplied path is a directory"""
        return self._getRPC(timeout=timeout).isDirectory(lfns)

    @checkCatalogArguments
    def getDirectorySize(self, lfns, longOut=False, fromFiles=False, timeout=120, recursiveSum=True):
        """Get the size of the supplied directory"""
        return self._getRPC(timeout=timeout).getDirectorySize(lfns, longOut, fromFiles, recursiveSum)

    ########################################################################
    #
    # Administrative database operations
    #

    def getCatalogCounters(self, timeout=120):
        """Get the number of registered directories, files and replicas in various tables"""
        return self._getRPC(timeout=timeout).getCatalogCounters()

    def rebuildDirectoryUsage(self, timeout=120):
        """Rebuild DirectoryUsage table from scratch"""
        return self._getRPC(timeout=timeout).rebuildDirectoryUsage()

    def repairCatalog(self, timeout=120):
        """Repair the catalog inconsistencies"""
        return self._getRPC(timeout=timeout).repairCatalog()

    ########################################################################
    # Metadata Catalog Operations
    #

    def addMetadataField(self, fieldName, fieldType, metaType="-d", timeout=120):
        """Add a new metadata field of the given type"""
        return self._getRPC(timeout=timeout).addMetadataField(fieldName, fieldType, metaType)

    def deleteMetadataField(self, fieldName, timeout=120):
        """Delete the metadata field"""
        return self._getRPC(timeout=timeout).deleteMetadataField(fieldName)

    def getMetadataFields(self, timeout=120):
        """Get all the metadata fields"""
        return self._getRPC(timeout=timeout).getMetadataFields()

    def setMetadata(self, path, metadatadict, timeout=120):
        """Set metadata parameter for the given path"""
        return self._getRPC(timeout=timeout).setMetadata(path, metadatadict)

    def setMetadataBulk(self, pathMetadataDict, timeout=120):
        """Set metadata parameter for the given path"""
        return self._getRPC(timeout=timeout).setMetadataBulk(pathMetadataDict)

    def removeMetadata(self, pathMetadataDict, timeout=120):
        """Remove the specified metadata for the given path"""
        return self._getRPC(timeout=timeout).removeMetadata(pathMetadataDict)

    def getDirectoryUserMetadata(self, path, timeout=120):
        """Get all the metadata valid for the given directory path"""
        return self._getRPC(timeout=timeout).getDirectoryUserMetadata(path)

    def findDirectoriesByMetadata(self, metaDict, path="/", timeout=120):
        """Find all the directories satisfying the given metadata set"""
        return self._getRPC(timeout=timeout).findDirectoriesByMetadata(metaDict, path)

    def getReplicasByMetadata(self, metaDict, path="/", allStatus=False, timeout=120):
        """Find all the files satisfying the given metadata set"""
        return self._getRPC(timeout=timeout).getReplicasByMetadata(metaDict, path, allStatus)

    def findFilesByMetadataDetailed(self, metaDict, path="/", timeout=120):
        """Find all the files satisfying the given metadata set"""
        return self._getRPC(timeout=timeout).findFilesByMetadataDetailed(metaDict, path)

    def findFilesByMetadataWeb(self, metaDict, path, startItem, maxItems, timeout=120):
        """Find files satisfying the given metadata set"""
        return self._getRPC(timeout=timeout).findFilesByMetadataWeb(metaDict, path, startItem, maxItems)

    def getCompatibleMetadata(self, metaDict, path="/", timeout=120):
        """Get metadata values compatible with the given metadata subset"""
        return self._getRPC(timeout=timeout).getCompatibleMetadata(metaDict, path)

    def addMetadataSet(self, setName, setDict, timeout=120):
        """Add a new metadata set"""
        return self._getRPC(timeout=timeout).addMetadataSet(setName, setDict)

    def getMetadataSet(self, setName, expandFlag, timeout=120):
        """Add a new metadata set"""
        return self._getRPC(timeout=timeout).getMetadataSet(setName, expandFlag)

    #########################################################################################
    #
    #  Dataset manipulation methods
    #

    @checkCatalogArguments
    def addDataset(self, datasets, timeout=120):
        """Add a new dynamic dataset defined by its meta query"""
        return self._getRPC(timeout=timeout).addDataset(datasets)

    @checkCatalogArguments
    def addDatasetAnnotation(self, datasetDict, timeout=120):
        """Add annotation to an already created dataset"""
        return self._getRPC(timeout=timeout).addDatasetAnnotation(datasetDict)

    @checkCatalogArguments
    def removeDataset(self, datasets, timeout=120):
        """Check the given dynamic dataset for changes since its definition"""
        return self._getRPC(timeout=timeout).removeDataset(datasets)

    @checkCatalogArguments
    def checkDataset(self, datasets, timeout=120):
        """Check the given dynamic dataset for changes since its definition"""
        return self._getRPC(timeout=timeout).checkDataset(datasets)

    @checkCatalogArguments
    def updateDataset(self, datasets, timeout=120):
        """Update the given dynamic dataset for changes since its definition"""
        return self._getRPC(timeout=timeout).updateDataset(datasets)

    @checkCatalogArguments
    def getDatasets(self, datasets, timeout=120):
        """Get parameters of the given dynamic dataset as they are stored in the database"""
        return self._getRPC(timeout=timeout).getDatasets(datasets)

    @checkCatalogArguments
    def getDatasetParameters(self, datasets, timeout=120):
        """Get parameters of the given dynamic dataset as they are stored in the database"""
        return self._getRPC(timeout=timeout).getDatasetParameters(datasets)

    @checkCatalogArguments
    def getDatasetAnnotation(self, datasets, timeout=120):
        """Get annotation of the given datasets"""
        return self._getRPC(timeout=timeout).getDatasetAnnotation(datasets)

    @checkCatalogArguments
    def freezeDataset(self, datasets, timeout=120):
        """Freeze the contents of the dataset making it effectively static"""
        return self._getRPC(timeout=timeout).freezeDataset(datasets)

    @checkCatalogArguments
    def releaseDataset(self, datasets, timeout=120):
        """Release the contents of the frozen dataset allowing changes in its contents"""
        return self._getRPC(timeout=timeout).releaseDataset(datasets)

    @checkCatalogArguments
    def getDatasetFiles(self, datasets, timeout=120):
        """Get lfns in the given dataset
        two lines !
        """
        return self._getRPC(timeout=timeout).getDatasetFiles(datasets)

    #############################################################################

    def getSEDump(self, seName, outputFilename):
        """
        Dump the content of an SE in the given file.
        The file contains a list of [lfn,checksum,size] dumped as csv,
        separated by '|'

        :param seName: name of the StorageElement
        :param outputFilename: path to the file where to dump it

        :returns: result from the TransferClient
        """

        dfc = TransferClient(self.serverURL)
        return dfc.receiveFile(outputFilename, seName)



#!/usr/bin/env python
"""
Find files in the FileCatalog using file metadata

Examples::

  $ dirac-dms-find-lfns Path=/lhcb/user "Size>1000" "CreationDate<2015-05-15"
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("", "Path=", "    Path to search for")
    Script.registerSwitch("", "SE=", "    (comma-separated list of) SEs/SE-groups to be searched")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        ["metaspec: metadata index specification (of the form: " '"meta=value" or "meta<value", "meta!=value", etc.)'],
        mandatory=False,
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    import DIRAC
    from DIRAC import gLogger
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from DIRAC.DataManagementSystem.Client.MetaQuery import MetaQuery, FILE_STANDARD_METAKEYS
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

    path = "/"
    seList = None
    for opt, val in Script.getUnprocessedSwitches():
        if opt == "Path":
            path = val
        elif opt == "SE":
            seList = resolveSEGroup(val.split(","))

    if seList:
        args.append("SE=%s" % ",".join(seList))
    fc = FileCatalog()
    result = fc.getMetadataFields()
    if not result["OK"]:
        gLogger.error("Can not access File Catalog:", result["Message"])
        DIRAC.exit(-1)
    typeDict = result["Value"]["FileMetaFields"]
    typeDict.update(result["Value"]["DirectoryMetaFields"])
    # Special meta tags
    typeDict.update(FILE_STANDARD_METAKEYS)

    if len(args) < 1:
        print("Error: No argument provided\n%s:" % Script.scriptName)
        gLogger.notice("MetaDataDictionary: \n%s" % str(typeDict))
        Script.showHelp(exitCode=1)

    mq = MetaQuery(typeDict=typeDict)
    result = mq.setMetaQuery(args)
    if not result["OK"]:
        gLogger.error("Illegal metaQuery:", result["Message"])
        DIRAC.exit(-1)
    metaDict = result["Value"]
    path = metaDict.pop("Path", path)

    result = fc.findFilesByMetadata(metaDict, path)
    if not result["OK"]:
        gLogger.error("Can not access File Catalog:", result["Message"])
        DIRAC.exit(-1)
    lfnList = sorted(result["Value"])

    gLogger.notice("\n".join(lfn for lfn in lfnList))


if __name__ == "__main__":
    main()
